# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM sets getter"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

import time

import FreeCAD

from femmesh import meshtools
from femtools.femutils import type_of_obj


class MeshSetsGetter:
    def __init__(
        self,
        analysis_obj,
        solver_obj,
        mesh_obj,
        member,
    ):
        # class attributes from parameter values
        self.analysis = analysis_obj
        self.solver_obj = solver_obj  # TODO without _obj
        self.mesh_object = mesh_obj  # TODO without _object
        self.member = member

        # more attributes
        self.analysis_type = self.solver_obj.AnalysisType
        self.document = self.analysis.Document
        self.fc_ver = FreeCAD.Version()
        self.ccx_nall = "Nall"
        self.ccx_eall = "Eall"
        self.ccx_evolumes = "Evolumes"
        self.ccx_efaces = "Efaces"
        self.ccx_eedges = "Eedges"
        self.mat_geo_sets = []
        self.theshape = None
        if self.mesh_object:
            if hasattr(self.mesh_object, "Shape"):
                self.theshape = self.mesh_object.Shape
            else:
                FreeCAD.Console.PrintWarning(
                    "A finite mesh without a link to a Shape was given. "
                    "Happen on pure mesh objects. "
                    "Not all methods do work without this link.\n"
                )
                # ATM only used in meshtools.get_femelement_direction1D_set
                # TODO somehow this is not smart, pur mesh objects might be used often
                if self.member.geos_beamsection and (
                    type_of_obj(self.solver_obj) == "Fem::SolverCcxTools"
                    or type_of_obj(self.solver_obj) == "Fem::SolverCalculiX"
                ):
                    FreeCAD.Console.PrintError(
                        "The mesh does not know the geometry it is made from. "
                        "Beam rotations can not retrieved but they are needed "
                        "for writing CalculiX solver input. "
                        "There might be problems in retrieving mesh data.\n"
                    )
                    # Z88 will run but CalculiX not
            self.femmesh = self.mesh_object.FemMesh
        else:
            FreeCAD.Console.PrintWarning(
                "No finite element mesh object was given to the writer class. "
                "In rare cases this might not be an error.\n"
            )
        self.femnodes_mesh = {}
        self.femelement_table = {}
        self.constraint_conflict_nodes = []
        self.femnodes_ele_table = {}
        self.femelements_edges_only = []
        self.femelements_faces_only = []
        self.femelement_volumes_table = {}
        self.femelement_faces_table = {}
        self.femelement_edges_table = {}
        self.femelement_count_test = True
        self.mat_geo_sets = []

        self._load_tables()

    def _load_tables(self):
        if self.mesh_object:
            if not self.femnodes_mesh:
                self.femnodes_mesh = self.femmesh.Nodes
            if not self.femelement_table:
                self.femelement_table = meshtools.get_femelement_table(self.femmesh)
            if not self.femnodes_ele_table:
                self.femnodes_ele_table = meshtools.get_femnodes_ele_table(
                    self.femnodes_mesh, self.femelement_table
                )

    # ********************************************************************************************
    # ********************************************************************************************
    # use set for node sets to be sure all nodes are unique
    # use sorted to be sure the order is the same on different runs
    # be aware a sorted set returns a list, because set are not sorted by default
    #     - done in return value of meshtools.get_femnodes_by_femobj_with_references
    # TODO FIXME might be appropriate for element sets and surfaceface sets too

    # ********************************************************************************************
    # ********************************************************************************************
    # get all known sets
    def get_mesh_sets(self):

        FreeCAD.Console.PrintMessage("\n")  # because of time print in separate line
        FreeCAD.Console.PrintMessage(
            "Get mesh data for constraints, materials and element geometry...\n"
        )
        FreeCAD.Console.PrintLog(
            "MeshSetsGetter: Get mesh data for "
            "node sets (groups), surface sets (groups) and element sets (groups)\n"
        )

        time_start = time.process_time()

        # materials and element geometry element sets getter
        self.get_element_sets_material_and_femelement_geometry()

        # constraints element sets getter
        self.get_constraints_centrif_elements()
        self.get_constraints_bodyheatsource_elements()

        # constraints node sets getter
        self.get_constraints_fixed_nodes()
        self.get_constraints_displacement_nodes()
        self.get_constraints_rigidbody_nodes()
        self.get_constraints_planerotation_nodes()

        # constraints surface sets getter
        self.get_constraints_contact_faces()
        self.get_constraints_tie_faces()
        self.get_constraints_sectionprint_faces()
        self.get_constraints_transform_nodes()
        self.get_constraints_temperature_nodes()
        self.get_constraints_initialtemperature_nodes()
        self.get_constraints_electrostatic_nodes()

        # constraints sets with constraint data
        self.get_constraints_force_nodeloads()
        self.get_constraints_pressure_faces()
        self.get_constraints_heatflux_faces()
        self.get_constraints_electrostatic_faces()
        self.get_constraints_electricchargedensity_faces()

        setstime = round((time.process_time() - time_start), 3)
        FreeCAD.Console.PrintMessage(f"Getting mesh data time: {setstime} seconds.\n")

    # ********************************************************************************************
    # ********************************************************************************************
    # node sets
    def get_constraints_fixed_nodes(self):
        if not self.member.cons_fixed:
            return
        # get nodes
        for femobj in self.member.cons_fixed:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)
            # add nodes to constraint_conflict_nodes, needed by constraint plane rotation
            for node in femobj["Nodes"]:
                self.constraint_conflict_nodes.append(node)
        # if mixed mesh with solids the node set needs to be split
        # because solid nodes do not have rotational degree of freedom
        if self.femmesh.Volumes and (
            len(self.member.geos_shellthickness) > 0 or len(self.member.geos_beamsection) > 0
        ):
            FreeCAD.Console.PrintMessage("We need to find the solid nodes.\n")
            if not self.femelement_volumes_table:
                self.femelement_volumes_table = meshtools.get_femelement_volumes_table(self.femmesh)
            for femobj in self.member.cons_fixed:
                # femobj --> dict, FreeCAD document object is femobj["Object"]
                nds_solid = []
                nds_faceedge = []
                for n in femobj["Nodes"]:
                    solid_node = False
                    for ve in self.femelement_volumes_table:
                        if n in self.femelement_volumes_table[ve]:
                            solid_node = True
                            nds_solid.append(n)
                            break
                    if not solid_node:
                        nds_faceedge.append(n)
                femobj["NodesSolid"] = set(nds_solid)
                femobj["NodesFaceEdge"] = set(nds_faceedge)

    def get_constraints_rigidbody_nodes(self):
        if not self.member.cons_rigidbody:
            return
        # get nodes
        for femobj in self.member.cons_rigidbody:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)
            # add nodes to constraint_conflict_nodes, needed by constraint plane rotation
            for node in femobj["Nodes"]:
                self.constraint_conflict_nodes.append(node)

    def get_constraints_displacement_nodes(self):
        if not self.member.cons_displacement:
            return
        # get nodes
        for femobj in self.member.cons_displacement:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)
            # add nodes to constraint_conflict_nodes, needed by constraint plane rotation
            for node in femobj["Nodes"]:
                self.constraint_conflict_nodes.append(node)

    def get_constraints_planerotation_nodes(self):
        if not self.member.cons_planerotation:
            return
        # get nodes
        for femobj in self.member.cons_planerotation:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)

    def get_constraints_transform_nodes(self):
        if not self.member.cons_transform:
            return
        # get nodes
        for femobj in self.member.cons_transform:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)

    def get_constraints_temperature_nodes(self):
        if not self.member.cons_temperature:
            return
        # get nodes
        for femobj in self.member.cons_temperature:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)

    def get_constraints_initialtemperature_nodes(self):
        if not self.member.cons_initialtemperature:
            return
        # get nodes
        for femobj in self.member.cons_initialtemperature:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)

    def get_constraints_fluidsection_nodes(self):
        if not self.member.geos_fluidsection:
            return
        # get nodes
        for femobj in self.member.geos_fluidsection:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"])
            femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(self.femmesh, femobj)

    def get_constraints_electrostatic_nodes(self):
        if not self.member.cons_electrostatic:
            return
        # get nodes
        for femobj in self.member.cons_electrostatic:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            if femobj["Object"].BoundaryCondition == "Dirichlet":
                print_obj_info(femobj["Object"])
                femobj["Nodes"] = meshtools.get_femnodes_by_femobj_with_references(
                    self.femmesh, femobj
                )

    def get_constraints_force_nodeloads(self):
        if not self.member.cons_force:
            return
        # check shape type of reference shape
        for femobj in self.member.cons_force:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            print_obj_info(femobj["Object"], log=True)
            if femobj["RefShapeType"] == "Vertex":
                FreeCAD.Console.PrintLog(
                    "    load on vertices --> The femelement_table "
                    "and femnodes_mesh are not needed for node load calculation.\n"
                )
            elif (
                femobj["RefShapeType"] == "Face"
                and meshtools.is_solid_femmesh(self.femmesh)
                and not meshtools.has_no_face_data(self.femmesh)
            ):
                FreeCAD.Console.PrintLog(
                    "    solid_mesh with face data --> The femelement_table is not "
                    "needed but the femnodes_mesh is needed for node load calculation.\n"
                )
                if not self.femnodes_mesh:
                    self.femnodes_mesh = self.femmesh.Nodes
            else:
                FreeCAD.Console.PrintLog(
                    "    mesh without needed data --> The femelement_table "
                    "and femnodes_mesh are not needed for node load calculation.\n"
                )
                if not self.femnodes_mesh:
                    self.femnodes_mesh = self.femmesh.Nodes
                if not self.femelement_table:
                    self.femelement_table = meshtools.get_femelement_table(self.femmesh)
        # get node loads
        FreeCAD.Console.PrintLog(
            "    Finite element mesh nodes will be retrieved by searching "
            "the appropriate nodes in the finite element mesh.\n"
        )
        FreeCAD.Console.PrintLog(
            "    The appropriate finite element mesh node load values will "
            "be calculated according to the finite element definition.\n"
        )
        for femobj in self.member.cons_force:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            frc_obj = femobj["Object"]
            print_obj_info(frc_obj)
            if frc_obj.Force == 0:
                FreeCAD.Console.PrintMessage("  Warning --> Force = 0\n")
            if femobj["RefShapeType"] == "Vertex":  # point load on vertices
                femobj["NodeLoadTable"] = meshtools.get_force_obj_vertex_nodeload_table(
                    self.femmesh, frc_obj
                )
            elif femobj["RefShapeType"] == "Edge":  # line load on edges
                femobj["NodeLoadTable"] = meshtools.get_force_obj_edge_nodeload_table(
                    self.femmesh, self.femelement_table, self.femnodes_mesh, frc_obj
                )
            elif femobj["RefShapeType"] == "Face":  # area load on faces
                femobj["NodeLoadTable"] = meshtools.get_force_obj_face_nodeload_table(
                    self.femmesh, self.femelement_table, self.femnodes_mesh, frc_obj
                )

    # ********************************************************************************************
    # ********************************************************************************************
    def _get_ccx_elements(self, obj):
        print_obj_info(obj)
        result = []
        ref_data = meshtools.pair_obj_reference(obj.References)
        for ref_pair in ref_data:
            result.append(meshtools.get_ccx_elements(self, ref_pair))

        return result

    # faces sets
    def get_constraints_pressure_faces(self):
        for femobj in self.member.cons_pressure:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            femobj["PressureFaces"] = result

    def get_constraints_electrostatic_faces(self):
        for femobj in self.member.cons_electrostatic:
            obj = femobj["Object"]
            if obj.BoundaryCondition == "Neumann":
                result = self._get_ccx_elements(obj)

                femobj["ElectricFluxFaces"] = result

    def get_constraints_electricchargedensity_faces(self):
        for femobj in self.member.cons_electricchargedensity:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            if obj.Mode in ["Interface", "Total Interface"]:
                femobj["ChargeDensityFaces"] = result
            elif obj.Mode in ["Source", "Total Source"]:
                femobj["ChargeDensityElements"] = result

    def get_constraints_contact_faces(self):
        for femobj in self.member.cons_contact:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            femobj["ContactSlaveFaces"] = result[:-1]
            femobj["ContactMasterFaces"] = result[-1:]

    # information in the regard of element faces constraints
    # forum post: https://forum.freecad.org/viewtopic.php?f=18&t=42783&p=370286#p366723
    # contact: master and slave could be the same face: rubber of a damper
    # tie: master and slave have to be separate faces AFA UR_ K
    # section print: only the element faces of solid elements
    #                from one side of the geometric face are needed

    def get_constraints_tie_faces(self):
        for femobj in self.member.cons_tie:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            femobj["TieSlaveFaces"] = result[:-1]
            femobj["TieMasterFaces"] = result[-1:]

    def get_constraints_sectionprint_faces(self):
        if not self.member.cons_sectionprint:
            return
        # TODO: use meshtools to get the surfaces
        # see constraint contact or constraint tie
        for femobj in self.member.cons_sectionprint:
            # femobj --> dict, FreeCAD document object is femobj["Object"]
            sectionprint_obj = femobj["Object"]
            if len(sectionprint_obj.References) > 1:
                FreeCAD.Console.PrintError(
                    "Only one reference shape allowed for a section print "
                    "but {} found: {}\n".format(
                        len(sectionprint_obj.References), sectionprint_obj.References
                    )
                )
            for o, elem_tup in sectionprint_obj.References:
                for elem in elem_tup:
                    # there should only be one reference for each section print object
                    # in the gui this is checked
                    ref_shape = o.Shape.getElement(elem)
                    if ref_shape.ShapeType == "Face":
                        v = self.mesh_object.FemMesh.getccxVolumesByFace(ref_shape)
                        if len(v) > 0:
                            femobj["SectionPrintFaces"] = v
                            # volume elements found
                            FreeCAD.Console.PrintLog(
                                "{}, surface {}, {} touching volume elements found\n".format(
                                    sectionprint_obj.Label, sectionprint_obj.Name, len(v)
                                )
                            )
                        else:
                            # no volume elements found, shell elements not allowed
                            FreeCAD.Console.PrintError(
                                "{}, surface {}, Error: "
                                "No volume elements found!\n".format(
                                    sectionprint_obj.Label, sectionprint_obj.Name
                                )
                            )
                    else:
                        # in Gui only Faces can be added
                        FreeCAD.Console.PrintError(
                            "Wrong reference shape type for {} "
                            "Only Faces are allowed, but a {} was found.\n".format(
                                sectionprint_obj.Name, ref_shape.ShapeType
                            )
                        )

    def get_constraints_heatflux_faces(self):
        for femobj in self.member.cons_heatflux:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            femobj["HeatFluxFaces"] = result

    # ********************************************************************************************
    # ********************************************************************************************
    # element sets constraints
    def get_constraints_centrif_elements(self):
        for femobj in self.member.cons_centrif:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            femobj["CentrifElements"] = result

    def get_constraints_bodyheatsource_elements(self):
        for femobj in self.member.cons_bodyheatsource:
            obj = femobj["Object"]
            result = self._get_ccx_elements(obj)

            femobj["BodyHeatSourceElements"] = result

    # ********************************************************************************************
    # ********************************************************************************************
    # element sets material and element geometry
    def get_solid_element_sets(self, femobjs):
        # get element ids and write them into the femobj
        all_found = False
        if self.femmesh.GroupCount:
            all_found = meshtools.get_femelement_sets_from_group_data(self.femmesh, femobjs)
            FreeCAD.Console.PrintMessage(all_found)
            FreeCAD.Console.PrintMessage("\n")
        if all_found is False:
            if not self.femelement_table:
                self.femelement_table = meshtools.get_femelement_table(self.femmesh)
            # we're going to use the binary search for get_femelements_by_femnodes()
            # thus we need the parameter values self.femnodes_ele_table
            if not self.femnodes_mesh:
                self.femnodes_mesh = self.femmesh.Nodes
            if not self.femnodes_ele_table:
                self.femnodes_ele_table = meshtools.get_femnodes_ele_table(
                    self.femnodes_mesh, self.femelement_table
                )
            control = meshtools.get_femelement_sets(
                self.femmesh, self.femelement_table, femobjs, self.femnodes_ele_table
            )
            # we only need to set it, if it is still True
            if (self.femelement_count_test is True) and (control is False):
                self.femelement_count_test = False

    def get_element_geometry2D_elements(self):
        # get element ids and write them into the objects
        FreeCAD.Console.PrintMessage("Shell thicknesses\n")
        if not self.femelement_faces_table:
            self.femelement_faces_table = meshtools.get_femelement_faces_table(self.femmesh)
        meshtools.get_femelement_sets(
            self.femmesh, self.femelement_faces_table, self.member.geos_shellthickness
        )

    def get_element_geometry1D_elements(self):
        # get element ids and write them into the objects
        FreeCAD.Console.PrintMessage("Beam sections\n")
        if not self.femelement_edges_table:
            self.femelement_edges_table = meshtools.get_femelement_edges_table(self.femmesh)
        meshtools.get_femelement_sets(
            self.femmesh, self.femelement_edges_table, self.member.geos_beamsection
        )

    def get_element_rotation1D_elements(self):
        # get for each geometry edge direction the element ids and rotation norma
        FreeCAD.Console.PrintMessage("Beam rotations\n")
        if self.theshape is None:
            FreeCAD.Console.PrintError(
                "Beam rotations set can not be retrieved, "
                "because the mesh does not know the Geometry it is made from\n"
            )
            return
        if not self.femelement_edges_table:
            self.femelement_edges_table = meshtools.get_femelement_edges_table(self.femmesh)
        meshtools.get_femelement_direction1D_set(
            self.femmesh, self.femelement_edges_table, self.member.geos_beamrotation, self.theshape
        )

    def get_element_fluid1D_elements(self):
        # get element ids and write them into the objects
        FreeCAD.Console.PrintMessage("Fluid sections\n")
        if not self.femelement_edges_table:
            self.femelement_edges_table = meshtools.get_femelement_edges_table(self.femmesh)
        meshtools.get_femelement_sets(
            self.femmesh, self.femelement_edges_table, self.member.geos_fluidsection
        )

    def get_material_elements(self):
        # it only works if either Volumes or Shellthicknesses or Beamsections
        # are in the material objects, it means it does not work
        # for mixed meshes and multiple materials, this is checked in check_prerequisites
        # the femelement_table is only calculated for
        # the highest dimension in get_femelement_table
        FreeCAD.Console.PrintMessage("Materials\n")
        if self.femmesh.Volumes:
            # we only could do this for volumes
            # if a mesh contains volumes we're going to use them in the analysis
            # but a mesh could contain
            # the element faces of the volumes as faces
            # and the edges of the faces as edges
            # there we have to check of some geometric objects
            # get element ids and write them into the femobj
            self.get_solid_element_sets(self.member.mats_linear)
        if self.member.geos_shellthickness:
            if not self.femelement_faces_table:
                self.femelement_faces_table = meshtools.get_femelement_faces_table(self.femmesh)
            meshtools.get_femelement_sets(
                self.femmesh, self.femelement_faces_table, self.member.mats_linear
            )
        if self.member.geos_beamsection or self.member.geos_fluidsection:
            if not self.femelement_edges_table:
                self.femelement_edges_table = meshtools.get_femelement_edges_table(self.femmesh)
            meshtools.get_femelement_sets(
                self.femmesh, self.femelement_edges_table, self.member.mats_linear
            )

    def get_element_sets_material_and_femelement_geometry(self):
        if not self.member.mats_linear:
            return
        # in any case if we have beams, we're going to need the element ids for the rotation elsets
        if self.member.geos_beamsection:
            # we will need to split the beam even for one beamobj
            # because no beam in z-direction can be used in ccx without a special adjustment
            # thus they need an own matgeoset
            self.get_element_rotation1D_elements()

        # get the element ids for face and edge elements and write them into the objects
        if len(self.member.geos_shellthickness) > 1:
            self.get_element_geometry2D_elements()
        if len(self.member.geos_beamsection) > 1:
            self.get_element_geometry1D_elements()
        if len(self.member.geos_fluidsection) > 1:
            self.get_element_fluid1D_elements()

        # get the element ids for material objects and write them into the material object
        if len(self.member.mats_linear) > 1:
            self.get_material_elements()

        # create the mat_geo_sets
        if len(self.member.mats_linear) == 1:
            if self.femmesh.Volumes:
                # we only could do this for volumes, if a mesh contains volumes
                # we're going to use them in the analysis
                # but a mesh could contain the element faces of the volumes as faces
                # and the edges of the faces as edges
                # there we have to check for some geometric objects
                self.get_mat_geo_sets_single_mat_solid()
            if len(self.member.geos_shellthickness) == 1:
                self.get_mat_geo_sets_single_mat_single_shell()
            elif len(self.member.geos_shellthickness) > 1:
                self.get_mat_geo_sets_single_mat_multiple_shell()
            if len(self.member.geos_beamsection) == 1:
                self.get_mat_geo_sets_single_mat_single_beam()
            elif len(self.member.geos_beamsection) > 1:
                self.get_mat_geo_sets_single_mat_multiple_beam()
            if len(self.member.geos_fluidsection) == 1:
                self.get_mat_geo_sets_single_mat_single_fluid()
            elif len(self.member.geos_fluidsection) > 1:
                self.get_mat_geo_sets_single_mat_multiple_fluid()
        elif len(self.member.mats_linear) > 1:
            if self.femmesh.Volumes:
                # we only could do this for volumes, if a mseh contains volumes
                # we're going to use them in the analysis
                # but a mesh could contain the element faces of the volumes as faces
                # and the edges of the faces as edges
                # there we have to check for some geometric objects
                # volume is a bit special
                # because retrieving ids from group mesh data is implemented
                self.get_mat_geo_sets_multiple_mat_solid()
            if len(self.member.geos_shellthickness) == 1:
                self.get_mat_geo_sets_multiple_mat_single_shell()
            elif len(self.member.geos_shellthickness) > 1:
                self.get_mat_geo_sets_multiple_mat_multiple_shell()
            if len(self.member.geos_beamsection) == 1:
                self.get_mat_geo_sets_multiple_mat_single_beam()
            elif len(self.member.geos_beamsection) > 1:
                self.get_mat_geo_sets_multiple_mat_multiple_beam()
            if len(self.member.geos_fluidsection) == 1:
                self.get_mat_geo_sets_multiple_mat_single_fluid()
            elif len(self.member.geos_fluidsection) > 1:
                self.get_mat_geo_sets_multiple_mat_multiple_fluid()

    # self.mat_geo_sets = [ {
    #                        "ccx_elset" : [e1, e2, e3, ... , en] or elements set name strings
    #                        "ccx_elset_name" : "ccx_identifier_elset"
    #                        "mat_obj_name" : "mat_obj.Name"
    #                        "ccx_mat_name" : "mat_obj.Material["Name"]"   !!! not unique !!!
    #                        "beamsection_obj" : "beamsection_obj"         if exists
    #                        "fluidsection_obj" : "fluidsection_obj"       if exists
    #                        "shellthickness_obj" : shellthickness_obj"    if exists
    #                        "beam_axis_m" : main local beam axis          for beams only
    #                     },
    #                     {}, ... , {} ]

    # beam
    # TODO support multiple beamrotations
    # we do not need any more any data from the rotation document object,
    # thus we do not need to save the rotation document object name in the else
    def get_mat_geo_sets_single_mat_single_beam(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        beamsec_obj = self.member.geos_beamsection[0]["Object"]
        beamrot_data = self.member.geos_beamrotation[0]
        for i, beamdirection in enumerate(beamrot_data["FEMRotations1D"]):
            # ID's for this direction
            elset_data = beamdirection["ids"]
            names = [
                {"short": "M0"},
                {"short": "B0"},
                {"short": beamrot_data["ShortName"]},
                {"short": "D" + str(i)},
            ]
            matgeoset = {}
            matgeoset["ccx_elset"] = elset_data
            matgeoset["ccx_elset_name"] = get_elset_name_short(names)
            matgeoset["mat_obj_name"] = mat_obj.Name
            matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
            matgeoset["beamsection_obj"] = beamsec_obj
            # beam_axis_m for this direction
            matgeoset["beam_axis_m"] = beamdirection["beam_axis_m"]
            self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_single_mat_multiple_beam(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        beamrot_data = self.member.geos_beamrotation[0]
        for beamsec_data in self.member.geos_beamsection:
            beamsec_obj = beamsec_data["Object"]
            beamsec_ids = set(beamsec_data["FEMElements"])
            for i, beamdirection in enumerate(beamrot_data["FEMRotations1D"]):
                beamdir_ids = set(beamdirection["ids"])
                # empty intersection sets possible
                elset_data = list(sorted(beamsec_ids.intersection(beamdir_ids)))
                if elset_data:
                    names = [
                        {"short": "M0"},
                        {"short": beamsec_data["ShortName"]},
                        {"short": beamrot_data["ShortName"]},
                        {"short": "D" + str(i)},
                    ]
                    matgeoset = {}
                    matgeoset["ccx_elset"] = elset_data
                    matgeoset["ccx_elset_name"] = get_elset_name_short(names)
                    matgeoset["mat_obj_name"] = mat_obj.Name
                    matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
                    matgeoset["beamsection_obj"] = beamsec_obj
                    # beam_axis_m for this direction
                    matgeoset["beam_axis_m"] = beamdirection["beam_axis_m"]
                    self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_multiple_mat_single_beam(self):
        beamsec_obj = self.member.geos_beamsection[0]["Object"]
        beamrot_data = self.member.geos_beamrotation[0]
        for mat_data in self.member.mats_linear:
            mat_obj = mat_data["Object"]
            mat_ids = set(mat_data["FEMElements"])
            for i, beamdirection in enumerate(beamrot_data["FEMRotations1D"]):
                beamdir_ids = set(beamdirection["ids"])
                elset_data = list(sorted(mat_ids.intersection(beamdir_ids)))
                if elset_data:
                    names = [
                        {"short": mat_data["ShortName"]},
                        {"short": "B0"},
                        {"short": beamrot_data["ShortName"]},
                        {"short": "D" + str(i)},
                    ]
                    matgeoset = {}
                    matgeoset["ccx_elset"] = elset_data
                    matgeoset["ccx_elset_name"] = get_elset_name_short(names)
                    matgeoset["mat_obj_name"] = mat_obj.Name
                    matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
                    matgeoset["beamsection_obj"] = beamsec_obj
                    # beam_axis_m for this direction
                    matgeoset["beam_axis_m"] = beamdirection["beam_axis_m"]
                    self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_multiple_mat_multiple_beam(self):
        beamrot_data = self.member.geos_beamrotation[0]
        for beamsec_data in self.member.geos_beamsection:
            beamsec_obj = beamsec_data["Object"]
            beamsec_ids = set(beamsec_data["FEMElements"])
            for mat_data in self.member.mats_linear:
                mat_obj = mat_data["Object"]
                mat_ids = set(mat_data["FEMElements"])
                for i, beamdirection in enumerate(beamrot_data["FEMRotations1D"]):
                    beamdir_ids = set(beamdirection["ids"])
                    # empty intersection sets possible
                    elset_data = list(
                        sorted(beamsec_ids.intersection(mat_ids).intersection(beamdir_ids))
                    )
                    if elset_data:
                        names = [
                            {"short": mat_data["ShortName"]},
                            {"short": beamsec_data["ShortName"]},
                            {"short": beamrot_data["ShortName"]},
                            {"short": "D" + str(i)},
                        ]
                        matgeoset = {}
                        matgeoset["ccx_elset"] = elset_data
                        matgeoset["ccx_elset_name"] = get_elset_name_short(names)
                        matgeoset["mat_obj_name"] = mat_obj.Name
                        matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
                        matgeoset["beamsection_obj"] = beamsec_obj
                        # beam_axis_m for this direction
                        matgeoset["beam_axis_m"] = beamdirection["beam_axis_m"]
                        self.mat_geo_sets.append(matgeoset)

    # fluid
    def get_mat_geo_sets_single_mat_single_fluid(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        fluidsec_obj = self.member.geos_fluidsection[0]["Object"]
        elset_data = self.ccx_eedges
        names = [{"short": "M0"}, {"short": "F0"}]
        matgeoset = {}
        matgeoset["ccx_elset"] = elset_data
        matgeoset["ccx_elset_name"] = get_elset_name_short(names)
        matgeoset["mat_obj_name"] = mat_obj.Name
        matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
        matgeoset["fluidsection_obj"] = fluidsec_obj
        self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_single_mat_multiple_fluid(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        for fluidsec_data in self.member.geos_fluidsection:
            fluidsec_obj = fluidsec_data["Object"]
            elset_data = fluidsec_data["FEMElements"]
            names = [{"short": "M0"}, {"short": fluidsec_data["ShortName"]}]
            matgeoset = {}
            matgeoset["ccx_elset"] = elset_data
            matgeoset["ccx_elset_name"] = get_elset_name_short(names)
            matgeoset["mat_obj_name"] = mat_obj.Name
            matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
            matgeoset["fluidsection_obj"] = fluidsec_obj
            self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_multiple_mat_single_fluid(self):
        fluidsec_obj = self.member.geos_fluidsection[0]["Object"]
        for mat_data in self.member.mats_linear:
            mat_obj = mat_data["Object"]
            elset_data = mat_data["FEMElements"]
            names = [{"short": mat_data["ShortName"]}, {"short": "F0"}]
            matgeoset = {}
            matgeoset["ccx_elset"] = elset_data
            matgeoset["ccx_elset_name"] = get_elset_name_short(names)
            matgeoset["mat_obj_name"] = mat_obj.Name
            matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
            matgeoset["fluidsection_obj"] = fluidsec_obj
            self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_multiple_mat_multiple_fluid(self):
        for fluidsec_data in self.member.geos_fluidsection:
            fluidsec_obj = fluidsec_data["Object"]
            for mat_data in self.member.mats_linear:
                mat_obj = mat_data["Object"]
                fluidsec_ids = set(fluidsec_data["FEMElements"])
                mat_ids = set(mat_data["FEMElements"])
                # empty intersection sets possible
                elset_data = list(sorted(fluidsec_ids.intersection(mat_ids)))
                if elset_data:
                    names = [
                        {"short": mat_data["ShortName"]},
                        {"short": fluidsec_data["ShortName"]},
                    ]
                    matgeoset = {}
                    matgeoset["ccx_elset"] = elset_data
                    matgeoset["ccx_elset_name"] = get_elset_name_short(names)
                    matgeoset["mat_obj_name"] = mat_obj.Name
                    matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
                    matgeoset["fluidsection_obj"] = fluidsec_obj
                    self.mat_geo_sets.append(matgeoset)

    # shell
    def get_mat_geo_sets_single_mat_single_shell(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        shellth_obj = self.member.geos_shellthickness[0]["Object"]
        elset_data = self.ccx_efaces
        names = [{"long": mat_obj.Name, "short": "M0"}, {"long": shellth_obj.Name, "short": "S0"}]
        matgeoset = {}
        matgeoset["ccx_elset"] = elset_data
        matgeoset["ccx_elset_name"] = get_elset_name_standard(names)
        matgeoset["mat_obj_name"] = mat_obj.Name
        matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
        matgeoset["shellthickness_obj"] = shellth_obj
        self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_single_mat_multiple_shell(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        for shellth_data in self.member.geos_shellthickness:
            shellth_obj = shellth_data["Object"]
            elset_data = shellth_data["FEMElements"]
            names = [
                {"long": mat_obj.Name, "short": "M0"},
                {"long": shellth_obj.Name, "short": shellth_data["ShortName"]},
            ]
            matgeoset = {}
            matgeoset["ccx_elset"] = elset_data
            matgeoset["ccx_elset_name"] = get_elset_name_standard(names)
            matgeoset["mat_obj_name"] = mat_obj.Name
            matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
            matgeoset["shellthickness_obj"] = shellth_obj
            self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_multiple_mat_single_shell(self):
        shellth_obj = self.member.geos_shellthickness[0]["Object"]
        for mat_data in self.member.mats_linear:
            mat_obj = mat_data["Object"]
            elset_data = mat_data["FEMElements"]
            names = [
                {"long": mat_obj.Name, "short": mat_data["ShortName"]},
                {"long": shellth_obj.Name, "short": "S0"},
            ]
            matgeoset = {}
            matgeoset["ccx_elset"] = elset_data
            matgeoset["ccx_elset_name"] = get_elset_name_standard(names)
            matgeoset["mat_obj_name"] = mat_obj.Name
            matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
            matgeoset["shellthickness_obj"] = shellth_obj
            self.mat_geo_sets.append(matgeoset)

    def get_mat_geo_sets_multiple_mat_multiple_shell(self):
        for shellth_data in self.member.geos_shellthickness:
            shellth_obj = shellth_data["Object"]
            for mat_data in self.member.mats_linear:
                mat_obj = mat_data["Object"]
                shellth_ids = set(shellth_data["FEMElements"])
                mat_ids = set(mat_data["FEMElements"])
                # empty intersection sets possible
                elset_data = list(sorted(shellth_ids.intersection(mat_ids)))
                if elset_data:
                    names = [
                        {"long": mat_obj.Name, "short": mat_data["ShortName"]},
                        {"long": shellth_obj.Name, "short": shellth_data["ShortName"]},
                    ]
                    matgeoset = {}
                    matgeoset["ccx_elset"] = elset_data
                    matgeoset["ccx_elset_name"] = get_elset_name_standard(names)
                    matgeoset["mat_obj_name"] = mat_obj.Name
                    matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
                    matgeoset["shellthickness_obj"] = shellth_obj
                    self.mat_geo_sets.append(matgeoset)

    # solid
    def get_mat_geo_sets_single_mat_solid(self):
        mat_obj = self.member.mats_linear[0]["Object"]
        elset_data = self.ccx_evolumes
        names = [{"long": mat_obj.Name, "short": "M0"}, {"long": "Solid", "short": "Solid"}]
        matgeoset = {}
        matgeoset["ccx_elset"] = elset_data
        matgeoset["ccx_elset_name"] = get_elset_name_standard(names)
        matgeoset["mat_obj_name"] = mat_obj.Name
        matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
        self.mat_geo_sets.append(matgeoset)
        print(self.mat_geo_sets)

    def get_mat_geo_sets_multiple_mat_solid(self):
        for mat_data in self.member.mats_linear:
            mat_obj = mat_data["Object"]
            elset_data = mat_data["FEMElements"]
            names = [
                {"long": mat_obj.Name, "short": mat_data["ShortName"]},
                {"long": "Solid", "short": "Solid"},
            ]
            matgeoset = {}
            matgeoset["ccx_elset"] = elset_data
            matgeoset["ccx_elset_name"] = get_elset_name_standard(names)
            matgeoset["mat_obj_name"] = mat_obj.Name
            matgeoset["ccx_mat_name"] = mat_obj.Material["Name"]
            self.mat_geo_sets.append(matgeoset)


# ************************************************************************************************
# Helpers


# ccx elset names:
# M .. Material
# B .. Beam
# R .. BeamRotation
# D .. Direction
# F .. Fluid
# S .. Shell,
# TODO write comment into input file to elset ids and elset attributes


def get_elset_name_standard(names):
    # standard max length = 80
    elset_name = ""
    for name in names:
        elset_name += name["long"]
    if len(elset_name) < 81:
        return elset_name
    else:
        elset_name = ""
        for name in names:
            elset_name += name["short"]
        if len(elset_name) < 81:
            return elset_name
        else:
            error = (
                "FEM: Trouble in elset name, because an "
                "elset name is longer than 80 character! {}\n".format(elset_name)
            )
            raise Exception(error)


def get_elset_name_short(names):
    # restricted max length = 20 (elsets)
    # in CalculiX solver input this is needed for beam elsets
    elset_name = ""
    for name in names:
        elset_name += name["short"]
    if len(elset_name) < 21:
        return elset_name
    else:
        error = (
            "FEM: Trouble in elset name, because an"
            "short elset name is longer than 20 characters! {}\n".format(elset_name)
        )
        raise Exception(error)


def print_obj_info(obj, log=False):
    if log is False:
        FreeCAD.Console.PrintMessage(f"{obj.Label}:\n")
        FreeCAD.Console.PrintMessage(f"    Type: {type_of_obj(obj)}, Name: {obj.Name}\n")
    else:
        FreeCAD.Console.PrintLog(f"{obj.Label}:\n")
        FreeCAD.Console.PrintLog(f"    Type: {type_of_obj(obj)}, Name: {obj.Name}\n")


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

from unittest import TestCase
from unittest.mock import patch

from PySideWrapper import QtCore, QtWidgets

from Addon import Addon
from addonmanager_update_all_gui import UpdateAllGUI
from AddonManagerTest.gui.gui_mocks import DialogWatcher, DialogInteractor
import addonmanager_freecad_interface as fci


class TestUpdateAllGUI(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_run_shows_dialog(self):
        # Arrange
        addons = [Addon("Test 1"), Addon("Test 2"), Addon("Test 3")]
        update_all_gui = UpdateAllGUI(addons)
        dialog_watcher = DialogWatcher(
            "AddonManager_UpdateAllDialog",
            QtWidgets.QDialogButtonBox.Close,
        )

        # Act
        update_all_gui.run()
        while dialog_watcher.timer.isActive():
            QtCore.QCoreApplication.processEvents()

        # Assert
        self.assertTrue(dialog_watcher.dialog_found)

    def test_run_shows_dialog_with_no_addons(self):
        # Arrange
        update_all_gui = UpdateAllGUI([])
        dialog_watcher = DialogWatcher(
            "AddonManager_UpdateAllDialog",
            QtWidgets.QDialogButtonBox.Close,
        )

        # Act
        update_all_gui.run()
        while dialog_watcher.timer.isActive():
            QtCore.QCoreApplication.processEvents()

        # Assert
        self.assertTrue(dialog_watcher.dialog_found)

    @patch("addonmanager_update_all_gui.MissingDependencies")
    @patch("addonmanager_update_all_gui.UpdateAllGUI.proceed")
    def test_run_calls_proceed_with_no_missing_deps(self, mock_proceed, _mock_missing_deps):
        # Obviously this needs to be fixed for real, but with several real errors to deal with, just
        # getting the rest of the CI to work is higher priority. -chennes, 11/30/2025
        self.skipTest(
            "This test is segfaulting in the CI even though it runs locally without issue"
        )
        # Arrange
        addons = [Addon("Test 1"), Addon("Test 2"), Addon("Test 3")]
        update_all_gui = UpdateAllGUI(addons)

        def update_button_clicker(window):
            buttons = window.findChildren(QtWidgets.QPushButton)
            for button in buttons:
                if button.text() == fci.translate("AddonsInstaller", "Update Selected Addons"):
                    button.click()
                    return

        dialog_interactor = DialogInteractor("AddonManager_UpdateAllDialog", update_button_clicker)

        # Act
        update_all_gui.run()
        while dialog_interactor.timer.isActive():
            QtCore.QCoreApplication.processEvents()

        # Assert
        mock_proceed.assert_called_once()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Defines the base class for Arch module unit tests."""

import unittest
import FreeCAD


class TestArchBase(unittest.TestCase):

    def setUp(self):
        """
        Set up a new, clean document for the test. Ensure test isolation by creating a
        uniquely-named document and cleaning up any potential leftovers from a previously failed
        run.
        """
        self.doc_name = self.__class__.__name__

        # Close any document of the same name that might have been left over from a crashed or
        # aborted test run. FreeCAD.getDocument() raises a NameError if the document is not found,
        # so we wrap this check in a try...except block.
        try:
            FreeCAD.getDocument(self.doc_name)
            # If getDocument() succeeds, the document exists and must be closed.
            FreeCAD.closeDocument(self.doc_name)
        except NameError:
            # This is the expected path on a clean run; do nothing.
            pass

        # Create a fresh document for the current test.
        self.document = FreeCAD.newDocument(self.doc_name)
        self.assertEqual(self.document.Name, self.doc_name)

    def tearDown(self):
        """Close the test document after all tests in the class are complete."""
        if hasattr(self, "document") and self.document:
            try:
                FreeCAD.closeDocument(self.document.Name)
            except Exception as e:
                FreeCAD.Console.PrintError(
                    f"Error during tearDown in {self.__class__.__name__}: {e}\n"
                )

    def printTestMessage(self, text, prepend_text="Test ", end="\n"):
        """Write messages to the console including the line ending.

        Messages will be prepended with "Test ", unless an empty string is
        passed as the prepend_text argument
        """
        FreeCAD.Console.PrintMessage(prepend_text + text + end)



# ***************************************************************************
# *   Copyright (c) 2020 Wilfried Hortschitz  <w.hortschitz@gmail.com>      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM solver Elmer equation object Electricforce"
__author__ = "Wilfried Hortschitz"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from femtools import femutils
from ... import equationbase
from . import linear

SOLVER_EXEC_METHODS = ["After Timestep", "Always"]


def create(doc, name="Electricforce"):
    return femutils.createObject(doc, name, Proxy, ViewProxy)


class Proxy(linear.Proxy, equationbase.ElectricforceProxy):

    Type = "Fem::EquationElmerElectricforce"

    def __init__(self, obj):
        super().__init__(obj)

        obj.addProperty(
            "App::PropertyEnumeration",
            "ExecSolver",
            "Electric Force",
            (
                "That solver is only executed after solution converged\n"
                "To execute always, change to 'Always'"
            ),
            locked=True,
        )

        obj.ExecSolver = SOLVER_EXEC_METHODS
        obj.ExecSolver = "After Timestep"
        # Electrostatic has priority 10 and Electricforce needs
        # the potential field calculated by Electrostatic
        # therefore set priority to 5
        obj.Priority = 5


class ViewProxy(linear.ViewProxy, equationbase.ElectricforceViewProxy):
    pass


##  @}



# FreeCAD TemplatePyMod module  
# (c) 2011 Werner Mayer LGPL

import FreeCAD as App
import FreeCADGui as Gui
from PySide import QtGui,QtCore

class MyLineEdit(QtGui.QLineEdit):
    pass

class TaskWatcher:
    def __init__(self):
        self.commands = ["Part_Box", "Part_Sphere", "Part_Cylinder"]
        self.title = "Create primitives"
        self.icon = "Part_Sphere"
        self.widgets = [MyLineEdit()]
        self.widgets[0].setText("Line edit inside task box")
    def shouldShow(self):
        return App.ActiveDocument is not None

class TaskLineEdit:
    def __init__(self):
        self.widgets = [MyLineEdit()]
        self.widgets[0].setText("Line edit with no task box")
    def shouldShow(self):
        return True

class TaskWatcherFilter:
    def __init__(self):
        self.commands = ["Sketcher_NewSketch", "PartDesign_Fillet", "PartDesign_Chamfer"]
        self.filter = "SELECT Part::Feature SUBELEMENT Face COUNT 1"
        self.title = "Face tools"
        self.icon = "Part_Box_Parametric"

class TaskPanel:
    def __init__(self):
        self.ui = App.getResourceDir() + "Mod/TemplatePyMod/TaskPanel.ui"

    def accept(self):
        return True

    def reject(self):
        return True

    def clicked(self, index):
        pass

    def open(self):
        pass

    def needsFullSpace(self):
        return False

    def isAllowedAlterSelection(self):
        return True

    def isAllowedAlterView(self):
        return True

    def isAllowedAlterDocument(self):
        return True

    def getStandardButtons(self):
        return QtGui.QDialogButtonBox.Ok

    def helpRequested(self):
        pass

    def setupUi(self):
        mw = self.getMainWindow()
        form = mw.findChild(QtGui.QWidget, "TaskPanel")
        form.pushButton = form.findChild(QtGui.QPushButton, "pushButton")
        form.listWidget = form.findChild(QtGui.QListWidget, "listWidget")
        self.form = form
        #Connect Signals and Slots
        QtCore.QObject.connect(form.pushButton, QtCore.SIGNAL("clicked()"), self.addElement)

    def getMainWindow(self):
        "returns the main window"
        # using QtGui.QApplication.activeWindow() isn't very reliable because if another
        # widget than the mainwindow is active (e.g. a dialog) the wrong widget is
        # returned
        toplevel = QtGui.QApplication.topLevelWidgets()
        for i in toplevel:
            if i.metaObject().className() == "Gui::MainWindow":
                return i
        raise RuntimeError("No main window found")

    def addElement(self):
        item=QtGui.QInputDialog.getText(self.form, 'Add item', 'Enter:')
        if item[1]:
            self.form.listWidget.addItem(item[0])

class TaskCalendar:
    def __init__(self):
        self.form = QtGui.QCalendarWidget()

class TaskManyTaskBoxes:
    "illustrates how to add several taskboxes"
    def __init__(self):
        widget1 = QtGui.QCalendarWidget()
        widget2 = QtGui.QWidget()
        widget2.setWindowTitle("My Test Box")
        text = QtGui.QLabel("testBox",widget2)
        text.setObjectName("label")
        self.form = [widget1,widget2]

def createTask():
    Gui.Control.addTaskWatcher([TaskWatcher(), TaskLineEdit(), TaskWatcherFilter()])
    panel = TaskCalendar()
    #panel = TaskPanel()
    Gui.Control.showDialog(panel)
    #panel.setupUi()
    return panel



# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Vector as vec

from BasicShapes import Shapes
from Draft import clone
from Part import makeLine

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Constraint Centrif",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["centrif", "fixed"],
        "solvers": ["ccxtools"],
        "material": "multimaterial",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.constraint_centrif import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=57770

constraint centrif, concerning CENTRIF label from ccx's *DLOAD card

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    # ring
    stiffener = doc.addObject("Part::Box", "Stiffener")
    stiffener.Length = 10
    stiffener.Width = 200
    stiffener.Height = 10
    stiffener.Placement.Base = (-5, -100, 0)
    circumference = Shapes.addTube(doc, "Circumference")
    circumference.Height = 10.0
    circumference.InnerRadius = 97.5
    circumference.OuterRadius = 102.5
    doc.recompute()

    fusion = doc.addObject("Part::MultiFuse", "Fusion")
    fusion.Shapes = [stiffener, circumference]
    fusion.Refine = True
    doc.recompute()

    centerhole = doc.addObject("Part::Cylinder", "CenterHole")
    centerhole.Radius = 3
    centerhole.Height = 20
    doc.recompute()

    ring_bottom = doc.addObject("Part::Cut", "RingBottom")
    ring_bottom.Base = fusion
    ring_bottom.Tool = centerhole
    ring_bottom.Refine = True
    doc.recompute()

    # standard ring
    ring_top = clone(ring_bottom, delta=vec(0, 0, 20))
    ring_top.Label = "RingTop"

    # compound of both rings
    geom_obj = doc.addObject("Part::Compound", "TheRingOfFire")
    geom_obj.Links = [ring_bottom, ring_top]
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # line for centrif axis
    sh_axis_line = makeLine(vec(0, 0, 0), vec(0, 0, 31))
    axis_line = doc.addObject("Part::Feature", "CentrifAxis")
    axis_line.Shape = sh_axis_line
    doc.recompute()
    if FreeCAD.GuiUp:
        axis_line.ViewObject.LineWidth = 5.0
        axis_line.ViewObject.LineColor = (1.0, 0.0, 0.0)

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.SplitInputWriter = False
    analysis.addObject(solver_obj)

    # materials
    material_obj_scotty = ObjectsFem.makeMaterialSolid(doc, "Steel_Scotty")
    mat = material_obj_scotty.Material
    mat["Name"] = "Steel_Scotty"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    mat["Density"] = "4000 kg/m^3"
    material_obj_scotty.Material = mat
    material_obj_scotty.References = [(ring_bottom, "Solid1")]
    analysis.addObject(material_obj_scotty)

    material_obj_std = ObjectsFem.makeMaterialSolid(doc, "Steel_Std")
    mat = material_obj_std.Material
    mat["Name"] = "Steel_Std"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    mat["Density"] = "8000 kg/m^3"
    material_obj_std.Material = mat
    material_obj_std.References = [(ring_top, "Solid1")]
    analysis.addObject(material_obj_std)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, ("Face6", "Face14"))]
    analysis.addObject(con_fixed)

    # constraint centrif
    con_centrif = ObjectsFem.makeConstraintCentrif(doc, "ConstraintCentrif")
    con_centrif.RotationFrequency = "100 Hz"
    con_centrif.RotationAxis = [(axis_line, "Edge1")]
    analysis.addObject(con_centrif)

    # mesh
    from .meshes.mesh_constraint_centrif_tetra10 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMax = "5.0 mm"

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2019 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import FreeCADGui
import Path
import Path.Base.Gui.Util as PathGuiUtil
import Path.Op.Gui.Base as PathOpGui
import Path.Op.Gui.CircularHoleBase as PathCircularHoleBaseGui
import Path.Op.ThreadMilling as PathThreadMilling
import PathGui
import PathScripts.PathUtils as PathUtils
import csv

from PySide.QtCore import QT_TRANSLATE_NOOP

from PySide import QtCore, QtGui


__title__ = "CAM Thread Milling Operation UI."
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "UI and Command for CAM Thread Milling Operation."

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


def fillThreads(form, dataFile, defaultSelect):
    form.threadName.blockSignals(True)
    select = form.threadName.currentText()
    Path.Log.debug("select = '{}'".format(select))
    form.threadName.clear()
    with open("{}Mod/CAM/Data/Threads/{}".format(FreeCAD.getHomePath(), dataFile)) as fp:
        reader = csv.DictReader(fp)
        for row in reader:
            form.threadName.addItem(row["name"], row)
    if select:
        form.threadName.setCurrentText(select)
    elif defaultSelect:
        form.threadName.setCurrentText(defaultSelect)
    form.threadName.setEnabled(True)
    form.threadName.blockSignals(False)


class TaskPanelOpPage(PathCircularHoleBaseGui.TaskPanelOpPage):
    """Controller for the thread milling operation's page"""

    def initPage(self, obj):
        self.majorDia = PathGuiUtil.QuantitySpinBox(self.form.threadMajor, obj, "MajorDiameter")
        self.minorDia = PathGuiUtil.QuantitySpinBox(self.form.threadMinor, obj, "MinorDiameter")
        self.pitch = PathGuiUtil.QuantitySpinBox(self.form.threadPitch, obj, "Pitch")

    def getForm(self):
        """getForm() ... return UI"""
        form = FreeCADGui.PySideUic.loadUi(":/panels/PageOpThreadMillingEdit.ui")
        comboToPropertyMap = [
            ("threadOrientation", "ThreadOrientation"),
            ("threadType", "ThreadType"),
            ("opDirection", "Direction"),
        ]
        enumTups = PathThreadMilling.ObjectThreadMilling.propertyEnumerations(dataType="raw")
        self.populateCombobox(form, enumTups, comboToPropertyMap)

        return form

    def getFields(self, obj):
        """getFields(obj) ... update obj's properties with values from the UI"""
        Path.Log.track()

        self.majorDia.updateProperty()
        self.minorDia.updateProperty()
        self.pitch.updateProperty()

        obj.ThreadOrientation = self.form.threadOrientation.currentData()
        obj.ThreadType = self.form.threadType.currentData()
        obj.ThreadName = self.form.threadName.currentText()
        obj.ThreadFit = self.form.threadFit.value()
        obj.Direction = self.form.opDirection.currentData()
        obj.Passes = self.form.opPasses.value()
        obj.LeadInOut = self.form.leadInOut.checkState() == QtCore.Qt.Checked
        obj.TPI = self.form.threadTPI.value()

        try:
            self.updateToolController(obj, self.form.toolController)
        except PathUtils.PathNoTCExistsException:
            title = translate("CAM", "No valid toolcontroller")
            message = translate(
                "CAM",
                "This operation requires a tool controller with a threadmilling tool",
            )

            self.show_error_message(title, message)

    def setFields(self, obj):
        """setFields(obj) ... update UI with obj properties' values"""
        Path.Log.track()

        self.selectInComboBox(obj.ThreadOrientation, self.form.threadOrientation)
        self.selectInComboBox(obj.ThreadType, self.form.threadType)
        self.selectInComboBox(obj.Direction, self.form.opDirection)

        self.form.threadName.blockSignals(True)
        self.form.threadName.setCurrentText(obj.ThreadName)
        self.form.threadName.blockSignals(False)
        self.form.threadFit.setValue(obj.ThreadFit)
        self.form.threadTPI.setValue(obj.TPI)

        self.form.opPasses.setValue(obj.Passes)
        self.form.leadInOut.setCheckState(
            QtCore.Qt.Checked if obj.LeadInOut else QtCore.Qt.Unchecked
        )

        self.majorDia.updateWidget()
        self.minorDia.updateWidget()
        self.pitch.updateWidget()

        self.setupToolController(obj, self.form.toolController)
        self._updateFromThreadType()

    def _isThreadCustom(self):
        return self.form.threadType.currentData() in [
            PathThreadMilling.ThreadTypeCustomInternal,
            PathThreadMilling.ThreadTypeCustomExternal,
        ]

    def _isThreadImperial(self):
        return self.form.threadType.currentData() in PathThreadMilling.ThreadTypesImperial

    def _isThreadMetric(self):
        return self.form.threadType.currentData() in PathThreadMilling.ThreadTypesMetric

    def _isThreadInternal(self):
        return self.form.threadType.currentData() in PathThreadMilling.ThreadTypesInternal

    def _isThreadExternal(self):
        return self.form.threadType.currentData() in PathThreadMilling.ThreadTypesExternal

    def _updateFromThreadType(self):

        if self._isThreadCustom():
            self.form.threadName.setEnabled(False)
            self.form.threadFit.setEnabled(False)
            self.form.threadFitLabel.setEnabled(False)
            self.form.threadPitch.setEnabled(True)
            self.form.threadPitchLabel.setEnabled(True)
            self.form.threadTPI.setEnabled(True)
            self.form.threadTPILabel.setEnabled(True)
        else:
            self.form.threadFit.setEnabled(True)
            self.form.threadFitLabel.setEnabled(True)
            if self._isThreadMetric():
                self.form.threadPitch.setEnabled(True)
                self.form.threadPitchLabel.setEnabled(True)
                self.form.threadTPI.setEnabled(False)
                self.form.threadTPILabel.setEnabled(False)
                self.form.threadTPI.setValue(0)
            else:
                self.form.threadPitch.setEnabled(False)
                self.form.threadPitchLabel.setEnabled(False)
                self.form.threadTPI.setEnabled(True)
                self.form.threadTPILabel.setEnabled(True)
                self.pitch.updateWidget(0)
            fillThreads(
                self.form,
                PathThreadMilling.ThreadTypeData[self.form.threadType.currentData()],
                self.obj.ThreadName,
            )
        self._updateFromThreadName()

    def _updateFromThreadName(self):
        if not self._isThreadCustom():
            thread = self.form.threadName.currentData()
            fit = float(self.form.threadFit.value()) / 100
            maxmin = float(thread["dMajorMin"])
            maxmax = float(thread["dMajorMax"])
            major = maxmin + (maxmax - maxmin) * fit
            minmin = float(thread["dMinorMin"])
            minmax = float(thread["dMinorMax"])
            minor = minmin + (minmax - minmin) * fit

            if self._isThreadMetric():
                pitch = float(thread["pitch"])
                self.pitch.updateWidget(pitch)

            if self._isThreadImperial():
                tpi = int(thread["tpi"])
                self.form.threadTPI.setValue(tpi)
                minor = minor * 25.4
                major = major * 25.4

            self.majorDia.updateWidget(major)
            self.minorDia.updateWidget(minor)

        self.setDirty()

    def getSignalsForUpdate(self, obj):
        """getSignalsForUpdate(obj) ... return list of signals which cause the receiver to update the model"""
        signals = []

        signals.append(self.form.threadMajor.editingFinished)
        signals.append(self.form.threadMinor.editingFinished)
        signals.append(self.form.threadPitch.editingFinished)
        signals.append(self.form.threadOrientation.currentIndexChanged)
        signals.append(self.form.threadTPI.editingFinished)
        signals.append(self.form.opDirection.currentIndexChanged)
        signals.append(self.form.opPasses.editingFinished)
        if hasattr(self.form.leadInOut, "checkStateChanged"):  # Qt version >= 6.7.0
            signals.append(self.form.leadInOut.checkStateChanged)
        else:  # Qt version < 6.7.0
            signals.append(self.form.leadInOut.stateChanged)

        signals.append(self.form.toolController.currentIndexChanged)

        return signals

    def registerSignalHandlers(self, obj):
        self.form.threadType.currentIndexChanged.connect(self._updateFromThreadType)
        self.form.threadName.currentIndexChanged.connect(self._updateFromThreadName)
        self.form.threadFit.valueChanged.connect(self._updateFromThreadName)


Command = PathOpGui.SetupOperation(
    "ThreadMilling",
    PathThreadMilling.Create,
    TaskPanelOpPage,
    "CAM_ThreadMilling",
    QT_TRANSLATE_NOOP("CAM_ThreadMilling", "Thread Milling"),
    QT_TRANSLATE_NOOP(
        "CAM_ThreadMilling",
        "Creates a Thread Milling toolpath from features of a base object",
    ),
    PathThreadMilling.SetupProperties,
)

FreeCAD.Console.PrintLog("Loading PathThreadMillingGui ... done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Carlo Pavan <carlopav@gmail.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for all annotation type objects.

This is used by many objects that show dimensions and text created on screen
through Coin (pivy).
- DimensionBase
- LinearDimension
- AngularDimension
- Label
- Text
"""
## @package draft_annotation
# \ingroup draftobjects
# \brief Provides the object code for all annotation type objects.

## \addtogroup draftobjects
# @{
from draftutils.messages import _log


class DraftAnnotation(object):
    """The Draft Annotation Base object.

    This class is not used directly, but inherited by all Draft annotation
    objects.

    LinearDimension through DimensionBase
    AngularDimension through DimensionBase
    Label
    Text
    """

    def onDocumentRestored(self, obj):
        """Execute code when the document is restored.

        Check if new properties are present after the object is restored
        in order to migrate older objects.
        """
        vobj = getattr(obj, "ViewObject", None)
        if vobj is None:
            return
        if not getattr(vobj, "Proxy", None):
            # Object was saved without GUI.
            # onDocumentRestored in the object class should restore the ViewObject.
            return

        if not hasattr(vobj, "ScaleMultiplier") or not hasattr(vobj, "AnnotationStyle"):
            self.add_missing_properties_0v19(obj, vobj)

        if hasattr(vobj, "ArrowType") or hasattr(vobj, "ArrowSize"):
            self.update_properties_1v1(obj, vobj)

    def add_missing_properties_0v19(self, obj, vobj):
        """Provide missing annotation properties."""
        multiplier = None
        if not hasattr(vobj, "ScaleMultiplier"):
            multiplier = 1.00
            _log("v0.19, " + obj.Name + ", added view property 'ScaleMultiplier'")
        if not hasattr(vobj, "AnnotationStyle"):
            _log("v0.19, " + obj.Name + ", added view property 'AnnotationStyle'")
        vobj.Proxy.set_annotation_properties(vobj, vobj.PropertiesList)
        if multiplier is not None:
            vobj.ScaleMultiplier = multiplier

    def update_properties_1v1(self, obj, vobj):
        """Update view properties."""
        vobj.Proxy.set_graphics_properties(vobj, vobj.PropertiesList)

        if hasattr(vobj, "ArrowType"):
            typ = obj.Proxy.Type
            if typ == "Label":
                vobj.ArrowTypeStart = vobj.ArrowType
            elif typ == "AngularDimension" or obj.Diameter or not vobj.Proxy.is_linked_to_circle():
                vobj.ArrowTypeStart = vobj.ArrowType
                vobj.ArrowTypeEnd = vobj.ArrowType
            else:  # Radial dimension
                vobj.ArrowTypeStart = "None"
                vobj.ArrowTypeEnd = vobj.ArrowType
            vobj.setPropertyStatus("ArrowType", "-LockDynamic")
            vobj.removeProperty("ArrowType")
        if hasattr(vobj, "ArrowSize"):
            vobj.ArrowSizeStart = vobj.ArrowSize
            if hasattr(vobj, "ArrowSizeEnd"):
                # Label objects do not have this property
                vobj.ArrowSizeEnd = vobj.ArrowSize
            vobj.setPropertyStatus("ArrowSize", "-LockDynamic")
            vobj.removeProperty("ArrowSize")
        _log("v1.1, " + obj.Name + ", migrated view properties")

    def dumps(self):

        return

    def loads(self, state):

        return


## @}



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM postprocessing table ViewProvider for the document object"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package view_post_table
#  \ingroup FEM
#  \brief view provider for post table object

import FreeCAD
import FreeCADGui

from PySide import QtGui, QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP

from . import view_base_fempostextractors
from . import view_base_fempostvisualization
from femtaskpanels import task_post_table
from femguiutils import vtk_table_view as vtv

from . import view_base_femobject

_GuiPropHelper = view_base_femobject._GuiPropHelper


class EditViewWidget(QtGui.QWidget):

    def __init__(self, obj, post_dialog):
        super().__init__()

        self._object = obj
        self._post_dialog = post_dialog

        # load the ui and set it up
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/PostTableFieldViewEdit.ui"
        )
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.__init_widget()

    def __init_widget(self):
        # set the other properties
        self.widget.Name.setText(self._object.ViewObject.Name)
        self.widget.Name.editingFinished.connect(self.legendChanged)

    @QtCore.Slot()
    def legendChanged(self):
        self._object.ViewObject.Name = self.widget.Name.text()


class EditFieldAppWidget(QtGui.QWidget):

    def __init__(self, obj, post_dialog):
        super().__init__()

        self._object = obj
        self._post_dialog = post_dialog

        # load the ui and set it up (we reuse histogram, as we need the exact same)
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/PostHistogramFieldAppEdit.ui"
        )
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.__init_widget()

    def __init_widget(self):
        # set the other properties

        self._post_dialog._enumPropertyToCombobox(self._object, "XField", self.widget.Field)
        self._post_dialog._enumPropertyToCombobox(self._object, "XComponent", self.widget.Component)
        self.widget.Extract.setChecked(self._object.ExtractFrames)

        self.widget.Field.activated.connect(self.fieldChanged)
        self.widget.Component.activated.connect(self.componentChanged)
        self.widget.Extract.toggled.connect(self.extractionChanged)

    @QtCore.Slot(int)
    def fieldChanged(self, index):
        self._object.XField = index
        self._post_dialog._enumPropertyToCombobox(self._object, "XComponent", self.widget.Component)
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def componentChanged(self, index):
        self._object.XComponent = index
        self._post_dialog._recompute()

    @QtCore.Slot(bool)
    def extractionChanged(self, extract):
        self._object.ExtractFrames = extract
        self._post_dialog._recompute()


class EditIndexAppWidget(QtGui.QWidget):

    def __init__(self, obj, post_dialog):
        super().__init__()

        self._object = obj
        self._post_dialog = post_dialog

        # load the ui and set it up (we reuse histogram, as we need the exact same)
        self.widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/PostHistogramIndexAppEdit.ui"
        )
        layout = QtGui.QVBoxLayout()
        layout.addWidget(self.widget)
        self.setLayout(layout)

        self.__init_widget()

    def __init_widget(self):
        # set the other properties

        self.widget.Index.setValue(self._object.Index)
        self._post_dialog._enumPropertyToCombobox(self._object, "XField", self.widget.Field)
        self._post_dialog._enumPropertyToCombobox(self._object, "XComponent", self.widget.Component)

        self.widget.Index.valueChanged.connect(self.indexChanged)
        self.widget.Field.activated.connect(self.fieldChanged)
        self.widget.Component.activated.connect(self.componentChanged)

        # sometimes weird sizes occur with spinboxes
        self.widget.Index.setMaximumHeight(self.widget.Field.sizeHint().height())

    @QtCore.Slot(int)
    def fieldChanged(self, index):
        self._object.XField = index
        self._post_dialog._enumPropertyToCombobox(self._object, "XComponent", self.widget.Component)
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def componentChanged(self, index):
        self._object.XComponent = index
        self._post_dialog._recompute()

    @QtCore.Slot(int)
    def indexChanged(self, value):
        self._object.Index = value
        self._post_dialog._recompute()


class VPPostTableFieldData(view_base_fempostextractors.VPPostExtractor):
    """
    A View Provider for extraction of 1D field data specially for tables
    """

    def __init__(self, vobj):
        super().__init__(vobj)

    def _get_properties(self):

        prop = [
            _GuiPropHelper(
                type="App::PropertyString",
                name="Name",
                group="Table",
                doc=QT_TRANSLATE_NOOP(
                    "FEM", "The name used in the table header. Default name is used if empty"
                ),
                value="",
            ),
        ]
        return super()._get_properties() + prop

    def attach(self, vobj):
        self.Object = vobj.Object
        self.ViewObject = vobj

    def getIcon(self):
        return ":/icons/FEM_PostField.svg"

    def get_app_edit_widget(self, post_dialog):
        return EditFieldAppWidget(self.Object, post_dialog)

    def get_view_edit_widget(self, post_dialog):
        return EditViewWidget(self.Object, post_dialog)

    def get_preview(self):
        name = QT_TRANSLATE_NOOP("FEM", "default")
        if self.ViewObject.Name:
            name = self.ViewObject.Name
        return (QtGui.QPixmap(), name)

    def get_default_color_property(self):
        return None


class VPPostTableIndexOverFrames(VPPostTableFieldData):
    """
    A View Provider for extraction of 1D index over frames data
    """

    def __init__(self, vobj):
        super().__init__(vobj)

    def getIcon(self):
        return ":/icons/FEM_PostIndex.svg"

    def get_app_edit_widget(self, post_dialog):
        return EditIndexAppWidget(self.Object, post_dialog)


class VPPostTable(view_base_fempostvisualization.VPPostVisualization):
    """
    A View Provider for Table plots
    """

    def __init__(self, vobj):
        super().__init__(vobj)

    def getIcon(self):
        return ":/icons/FEM_PostSpreadsheet.svg"

    def setEdit(self, vobj, mode):

        # build up the task panel
        taskd = task_post_table._TaskPanel(vobj)

        # show it
        FreeCADGui.Control.showDialog(taskd)

        return True

    def show_visualization(self):

        if not hasattr(self, "_tableview") or not self._tableview:
            main = FreeCADGui.getMainWindow()
            self._tableModel = vtv.VtkTableModel()
            self._tableview = vtv.VtkTableView(self._tableModel)
            self._tableview.setWindowTitle(self.Object.Label)
            self._tableview.setParent(main)
            self._tableview.setWindowFlags(QtGui.Qt.Dialog)
            self._tableview.resize(
                main.size().height() / 2, main.size().height() / 3
            )  # keep the aspect ratio

            self.update_visualization()

        self._tableview.show()

    def update_visualization(self):

        if not hasattr(self, "_tableModel") or not self._tableModel:
            return

        # we collect the header names from the viewproviders
        table = self.Object.Table
        header = {}
        for child in self.Object.Group:

            if not child.Source:
                continue

            new_name = child.ViewObject.Name
            if new_name:
                # this child uses a custom name. We try to find all
                # columns that are from this child and use custom header for it
                for i in range(table.GetNumberOfColumns()):
                    if child.Source.Name in table.GetColumnName(i):
                        header[table.GetColumnName(i)] = new_name

        self._tableModel.setTable(self.Object.Table, header)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2016 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

__title__ = "Arch Pipe tools"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

## @package ArchPipe
#  \ingroup ARCH
#  \brief The Pipe object and tools
#
#  This module provides tools to build Pipe and Pipe connector objects.
#  Pipes are tubular objects extruded along a base line.

import FreeCAD
import ArchComponent
import ArchIFC

from draftutils import params

if FreeCAD.GuiUp:
    from PySide.QtCore import QT_TRANSLATE_NOOP
    import FreeCADGui
    import Arch_rc
    from draftutils.translate import translate
else:
    # \cond
    def translate(ctxt, txt):
        return txt

    def QT_TRANSLATE_NOOP(ctxt, txt):
        return txt

    # \endcond


class _ArchPipe(ArchComponent.Component):
    "the Arch Pipe object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.Type = "Pipe"
        self.setProperties(obj)
        # IfcPipeSegment is new in IFC4
        from ArchIFC import IfcTypes

        if "Pipe Segment" in IfcTypes:
            obj.IfcType = "Pipe Segment"
        else:
            # IFC2x3 does not know a Pipe Segment
            obj.IfcType = "Building Element Proxy"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Diameter" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Diameter",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The diameter of this pipe, if not based on a profile"
                ),
                locked=True,
            )
        if not "Width" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Width",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The width of this pipe, if not based on a profile"
                ),
                locked=True,
            )
            obj.setPropertyStatus("Width", "Hidden")
        if not "Height" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Height",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The height of this pipe, if not based on a profile"
                ),
                locked=True,
            )
            obj.setPropertyStatus("Height", "Hidden")
        if not "Length" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Length",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The length of this pipe, if not based on an edge"
                ),
                locked=True,
            )
        if not "Profile" in pl:
            obj.addProperty(
                "App::PropertyLink",
                "Profile",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property", "An optional closed profile to base this pipe on"
                ),
                locked=True,
            )
        if not "OffsetStart" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "OffsetStart",
                "Pipe",
                QT_TRANSLATE_NOOP("App::Property", "Offset from the start point"),
                locked=True,
            )
        if not "OffsetEnd" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "OffsetEnd",
                "Pipe",
                QT_TRANSLATE_NOOP("App::Property", "Offset from the end point"),
                locked=True,
            )
        if not "WallThickness" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "WallThickness",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property", "The wall thickness of this pipe, if not based on a profile"
                ),
                locked=True,
            )
        if not "ProfileType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "ProfileType",
                "Pipe",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "If not based on a profile, this controls the profile of this pipe",
                ),
                locked=True,
            )
            obj.ProfileType = ["Circle", "Square", "Rectangle"]

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)
        if (
            obj.ProfileType == "Rectangle"
            and FreeCAD.ActiveDocument.getProgramVersion().split()[0] <= "1.1R42474"
        ):
            # in older versions Height and Width are inverted
            obj.Height, obj.Width = obj.Width, obj.Height
            FreeCAD.ActiveDocument.recompute()
            FreeCAD.Console.PrintWarning(
                "v1.1, "
                + obj.Label
                + ", "
                + translate("Arch", "corrected 'Height' and 'Width' properties")
                + "\n"
            )

    def loads(self, state):

        self.Type = "Pipe"

    def onChanged(self, obj, prop):
        if prop == "IfcType":
            root = ArchIFC.IfcProduct()
            root.setupIfcAttributes(obj)
            root.setupIfcComplexAttributes(obj)
        elif prop == "ProfileType":
            if obj.ProfileType == "Square":
                obj.setPropertyStatus("Height", "Hidden")
                obj.setPropertyStatus("Diameter", "Hidden")
                obj.setPropertyStatus("Width", "-Hidden")
            elif obj.ProfileType == "Rectangle":
                obj.setPropertyStatus("Height", "-Hidden")
                obj.setPropertyStatus("Diameter", "Hidden")
                obj.setPropertyStatus("Width", "-Hidden")
            else:
                obj.setPropertyStatus("Height", "Hidden")
                obj.setPropertyStatus("Diameter", "-Hidden")
                obj.setPropertyStatus("Width", "Hidden")

    def execute(self, obj):

        import math
        import Part
        import DraftGeomUtils

        if self.clone(obj):
            return
        pl = obj.Placement
        w = self.getWire(obj)
        if not w:
            FreeCAD.Console.PrintError(translate("Arch", "Unable to build the base path") + "\n")
            return
        if obj.OffsetStart.Value:
            e = w.Edges[0]
            v = e.Vertexes[-1].Point.sub(e.Vertexes[0].Point).normalize()
            v.multiply(obj.OffsetStart.Value)
            e = Part.LineSegment(e.Vertexes[0].Point.add(v), e.Vertexes[-1].Point).toShape()
            w = Part.Wire([e] + w.Edges[1:])
        if obj.OffsetEnd.Value:
            e = w.Edges[-1]
            v = e.Vertexes[0].Point.sub(e.Vertexes[-1].Point).normalize()
            v.multiply(obj.OffsetEnd.Value)
            e = Part.LineSegment(e.Vertexes[-1].Point.add(v), e.Vertexes[0].Point).toShape()
            w = Part.Wire(w.Edges[:-1] + [e])
        p = self.getProfile(obj)
        if not p:
            FreeCAD.Console.PrintError(translate("Arch", "Unable to build the profile") + "\n")
            return
        # move and rotate the profile to the first point
        if hasattr(p, "CenterOfMass"):
            c = p.CenterOfMass
        else:
            c = p.BoundBox.Center
        delta = w.Vertexes[0].Point - c
        p.translate(delta)
        import Draft

        if Draft.getType(obj.Base) == "BezCurve":
            v1 = obj.Base.Placement.multVec(obj.Base.Points[1]) - w.Vertexes[0].Point
        else:
            v1 = w.Vertexes[1].Point - w.Vertexes[0].Point
        # v2 = DraftGeomUtils.getNormal(p)
        # rot = FreeCAD.Rotation(v2,v1)
        # rotate keeping up vector
        if v1.getAngle(FreeCAD.Vector(0, 0, 1)) > 0.01:
            up = FreeCAD.Vector(0, 0, 1)
        else:
            up = FreeCAD.Vector(0, 1, 0)
        v1y = up.cross(v1)
        v1x = v1.cross(v1y)
        rot = FreeCAD.Rotation(v1x, v1y, v1, "ZYX")
        p.rotate(w.Vertexes[0].Point, rot.Axis, math.degrees(rot.Angle))
        p.rotate(w.Vertexes[0].Point, v1, 90)
        shapes = []
        try:
            if p.Faces:
                for f in p.Faces:
                    sh = w.makePipeShell([f.OuterWire], True, False, 2)
                    for shw in f.Wires:
                        if shw.hashCode() != f.OuterWire.hashCode():
                            sh2 = w.makePipeShell([shw], True, False, 2)
                            sh = sh.cut(sh2)
                    shapes.append(sh)
            elif p.Wires:
                for pw in p.Wires:
                    sh = w.makePipeShell([pw], True, False, 2)
                    shapes.append(sh)
        except Exception:
            FreeCAD.Console.PrintError(translate("Arch", "Unable to build the pipe") + "\n")
        else:
            if len(shapes) == 0:
                return
            elif len(shapes) == 1:
                sh = shapes[0]
            else:
                sh = Part.makeCompound(shapes)
            obj.Shape = self.processSubShapes(obj, sh, pl)
            if obj.Base:
                obj.Length = w.Length
            else:
                obj.Placement = pl

    def getWire(self, obj):

        import Part

        if obj.Base:
            if not hasattr(obj.Base, "Shape"):
                FreeCAD.Console.PrintError(
                    translate("Arch", "The base object is not a Part") + "\n"
                )
                return
            if len(obj.Base.Shape.Wires) != 1:
                FreeCAD.Console.PrintError(
                    translate("Arch", "Too many wires in the base shape") + "\n"
                )
                return
            if obj.Base.Shape.Wires[0].isClosed():
                FreeCAD.Console.PrintError(translate("Arch", "The base wire is closed") + "\n")
                return
            w = obj.Base.Shape.Wires[0]
        else:
            if obj.Length.Value == 0:
                return
            w = Part.Wire(
                [
                    Part.LineSegment(
                        FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(0, 0, obj.Length.Value)
                    ).toShape()
                ]
            )
        return w

    def getProfile(self, obj):

        import Part

        if obj.Profile:
            if not obj.Profile.getLinkedObject().isDerivedFrom("Part::Part2DObject"):
                FreeCAD.Console.PrintError(translate("Arch", "The profile is not a 2D Part") + "\n")
                return
            if not obj.Profile.Shape.Wires[0].isClosed():
                FreeCAD.Console.PrintError(translate("Arch", "The profile is not closed") + "\n")
                return
            p = obj.Profile.Shape.Wires[0]
        else:
            if obj.ProfileType == "Square":
                if obj.Width.Value == 0:
                    return
                p = Part.makePlane(
                    obj.Width.Value,
                    obj.Width.Value,
                    FreeCAD.Vector(-obj.Width.Value / 2, -obj.Width.Value / 2, 0),
                )
                if obj.WallThickness.Value and (obj.WallThickness.Value < obj.Width.Value / 2):
                    p2 = Part.makePlane(
                        obj.Width.Value - obj.WallThickness.Value * 2,
                        obj.Width.Value - obj.WallThickness.Value * 2,
                        FreeCAD.Vector(
                            obj.WallThickness.Value - obj.Width.Value / 2,
                            obj.WallThickness.Value - obj.Width.Value / 2,
                            0,
                        ),
                    )
                    p = p.cut(p2)
            elif obj.ProfileType == "Rectangle":
                if obj.Width.Value == 0:
                    return
                if obj.Height.Value == 0:
                    return
                p = Part.makePlane(
                    obj.Width.Value,
                    obj.Height.Value,
                    FreeCAD.Vector(-obj.Width.Value / 2, -obj.Height.Value / 2, 0),
                )
                if (
                    obj.WallThickness.Value
                    and (obj.WallThickness.Value < obj.Height.Value / 2)
                    and (obj.WallThickness.Value < obj.Width.Value / 2)
                ):
                    p2 = Part.makePlane(
                        obj.Width.Value - obj.WallThickness.Value * 2,
                        obj.Height.Value - obj.WallThickness.Value * 2,
                        FreeCAD.Vector(
                            obj.WallThickness.Value - obj.Width.Value / 2,
                            obj.WallThickness.Value - obj.Height.Value / 2,
                            0,
                        ),
                    )
                    p = p.cut(p2)
            else:
                if obj.Diameter.Value == 0:
                    return
                p = Part.Wire(
                    [
                        Part.Circle(
                            FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(0, 0, 1), obj.Diameter.Value / 2
                        ).toShape()
                    ]
                )
                if obj.WallThickness.Value and (obj.WallThickness.Value < obj.Diameter.Value / 2):
                    p2 = Part.Wire(
                        [
                            Part.Circle(
                                FreeCAD.Vector(0, 0, 0),
                                FreeCAD.Vector(0, 0, 1),
                                (obj.Diameter.Value / 2 - obj.WallThickness.Value),
                            ).toShape()
                        ]
                    )
                    p = Part.Face(p)
                    p2 = Part.Face(p2)
                    p = p.cut(p2)
        return p


class _ViewProviderPipe(ArchComponent.ViewProviderComponent):
    "A View Provider for the Pipe object"

    def __init__(self, vobj):

        ArchComponent.ViewProviderComponent.__init__(self, vobj)

    def getIcon(self):

        import Arch_rc

        return ":/icons/Arch_Pipe_Tree.svg"


class _ArchPipeConnector(ArchComponent.Component):
    "the Arch Pipe Connector object"

    def __init__(self, obj):

        ArchComponent.Component.__init__(self, obj)
        self.setProperties(obj)
        obj.IfcType = "Pipe Fitting"

    def setProperties(self, obj):

        pl = obj.PropertiesList
        if not "Radius" in pl:
            obj.addProperty(
                "App::PropertyLength",
                "Radius",
                "PipeConnector",
                QT_TRANSLATE_NOOP("App::Property", "The curvature radius of this connector"),
                locked=True,
            )
        if not "Pipes" in pl:
            obj.addProperty(
                "App::PropertyLinkList",
                "Pipes",
                "PipeConnector",
                QT_TRANSLATE_NOOP("App::Property", "The pipes linked by this connector"),
                locked=True,
            )
        if not "ConnectorType" in pl:
            obj.addProperty(
                "App::PropertyEnumeration",
                "ConnectorType",
                "PipeConnector",
                QT_TRANSLATE_NOOP("App::Property", "The type of this connector"),
                locked=True,
            )
            obj.ConnectorType = ["Corner", "Tee"]
            obj.setEditorMode("ConnectorType", 1)
        self.Type = "PipeConnector"

    def onDocumentRestored(self, obj):

        ArchComponent.Component.onDocumentRestored(self, obj)
        self.setProperties(obj)

    def execute(self, obj):

        if self.clone(obj):
            return

        tol = 1  # tolerance for alignment. This is only visual, we can keep it low...
        ptol = 0.001  # tolerance for coincident points

        import math
        import Part
        import DraftGeomUtils
        import ArchCommands

        if len(obj.Pipes) < 2:
            return
        if len(obj.Pipes) > 3:
            FreeCAD.Console.PrintWarning(
                translate("Arch", "Only the 3 first wires will be connected") + "\n"
            )
        if obj.Radius.Value == 0:
            return
        wires = []
        order = []
        for o in obj.Pipes:
            wires.append(o.Proxy.getWire(o))
        if wires[0].Vertexes[0].Point.sub(wires[1].Vertexes[0].Point).Length <= ptol:
            order = ["start", "start"]
            point = wires[0].Vertexes[0].Point
        elif wires[0].Vertexes[0].Point.sub(wires[1].Vertexes[-1].Point).Length <= ptol:
            order = ["start", "end"]
            point = wires[0].Vertexes[0].Point
        elif wires[0].Vertexes[-1].Point.sub(wires[1].Vertexes[-1].Point).Length <= ptol:
            order = ["end", "end"]
            point = wires[0].Vertexes[-1].Point
        elif wires[0].Vertexes[-1].Point.sub(wires[1].Vertexes[0].Point).Length <= ptol:
            order = ["end", "start"]
            point = wires[0].Vertexes[-1].Point
        else:
            FreeCAD.Console.PrintError(translate("Arch", "Common vertex not found") + "\n")
            return
        if order[0] == "start":
            v1 = wires[0].Vertexes[1].Point.sub(wires[0].Vertexes[0].Point).normalize()
        else:
            v1 = wires[0].Vertexes[-2].Point.sub(wires[0].Vertexes[-1].Point).normalize()
        if order[1] == "start":
            v2 = wires[1].Vertexes[1].Point.sub(wires[1].Vertexes[0].Point).normalize()
        else:
            v2 = wires[1].Vertexes[-2].Point.sub(wires[1].Vertexes[-1].Point).normalize()
        p = obj.Pipes[0].Proxy.getProfile(obj.Pipes[0])
        if not p:
            return
        # If the pipe has a non-zero WallThickness p is a shape instead of a wire:
        if p.ShapeType != "Wire":
            p = p.Wires
        p = Part.Face(p)
        if len(obj.Pipes) == 2:
            if obj.ConnectorType != "Corner":
                obj.ConnectorType = "Corner"
            if round(v1.getAngle(v2), tol) in [0, round(math.pi, tol)]:
                FreeCAD.Console.PrintError(translate("Arch", "Pipes are already aligned") + "\n")
                return
            normal = v2.cross(v1)
            offset = math.tan(math.pi / 2 - v1.getAngle(v2) / 2) * obj.Radius.Value
            v1.multiply(offset)
            v2.multiply(offset)
            self.setOffset(obj.Pipes[0], order[0], offset)
            self.setOffset(obj.Pipes[1], order[1], offset)
            # find center
            perp = v1.cross(normal).normalize()
            perp.multiply(obj.Radius.Value)
            center = point.add(v1).add(perp)
            # move and rotate the profile to the first point
            delta = point.add(v1) - p.CenterOfMass
            p.translate(delta)
            # vp = DraftGeomUtils.getNormal(p)
            # rot = FreeCAD.Rotation(vp,v1)
            # rotate keeping up vector
            if v1.getAngle(FreeCAD.Vector(0, 0, 1)) > 0.01:
                up = FreeCAD.Vector(0, 0, 1)
            else:
                up = FreeCAD.Vector(0, 1, 0)
            v1y = up.cross(v1)
            v1x = v1.cross(v1y)
            rot = FreeCAD.Rotation(v1x, v1y, v1, "ZYX")
            p.rotate(p.CenterOfMass, rot.Axis, math.degrees(rot.Angle))
            p.rotate(p.CenterOfMass, v1, 90)
            try:
                sh = p.revolve(center, normal, math.degrees(math.pi - v1.getAngle(v2)))
            except:
                FreeCAD.Console.PrintError(
                    translate("Arch", "Unable to revolve this connector") + "\n"
                )
                return
            # sh = Part.makeCompound([sh]+[Part.Vertex(point),Part.Vertex(point.add(v1)),Part.Vertex(center),Part.Vertex(point.add(v2))])
        else:
            if obj.ConnectorType != "Tee":
                obj.ConnectorType = "Tee"
            if wires[2].Vertexes[0].Point == point:
                order.append("start")
            elif wires[0].Vertexes[-1].Point == point:
                order.append("end")
            else:
                FreeCAD.Console.PrintError(translate("Arch", "Common vertex not found") + "\n")
            if order[2] == "start":
                v3 = wires[2].Vertexes[1].Point.sub(wires[2].Vertexes[0].Point).normalize()
            else:
                v3 = wires[2].Vertexes[-2].Point.sub(wires[2].Vertexes[-1].Point).normalize()
            if round(v1.getAngle(v2), tol) in [0, round(math.pi, tol)]:
                pair = [v1, v2, v3]
            elif round(v1.getAngle(v3), tol) in [0, round(math.pi, tol)]:
                pair = [v1, v3, v2]
            elif round(v2.getAngle(v3), tol) in [0, round(math.pi, tol)]:
                pair = [v2, v3, v1]
            else:
                FreeCAD.Console.PrintError(translate("Arch", "At least 2 pipes must align") + "\n")
                return
            offset = obj.Radius.Value
            v1.multiply(offset)
            v2.multiply(offset)
            v3.multiply(offset)
            self.setOffset(obj.Pipes[0], order[0], offset)
            self.setOffset(obj.Pipes[1], order[1], offset)
            self.setOffset(obj.Pipes[2], order[2], offset)
            normal = pair[0].cross(pair[2])
            # move and rotate the profile to the first point
            delta = point.add(pair[0]) - p.CenterOfMass
            p.translate(delta)
            vp = DraftGeomUtils.getNormal(p)
            rot = FreeCAD.Rotation(vp, pair[0])
            p.rotate(p.CenterOfMass, rot.Axis, math.degrees(rot.Angle))
            t1 = p.extrude(pair[1].multiply(2))
            # move and rotate the profile to the second point
            delta = point.add(pair[2]) - p.CenterOfMass
            p.translate(delta)
            vp = DraftGeomUtils.getNormal(p)
            rot = FreeCAD.Rotation(vp, pair[2])
            p.rotate(p.CenterOfMass, rot.Axis, math.degrees(rot.Angle))
            t2 = p.extrude(pair[2].negative().multiply(2))
            # create a cut plane
            cp = Part.makePolygon([point, point.add(pair[0]), point.add(normal), point])
            cp = Part.Face(cp)
            if cp.normalAt(0, 0).getAngle(pair[2]) < math.pi / 2:
                cp.reverse()
            cf, cv, invcv = ArchCommands.getCutVolume(cp, t2)
            t2 = t2.cut(cv)
            sh = t1.fuse(t2)
        obj.Shape = sh

    def setOffset(self, pipe, pos, offset):

        if pos == "start":
            if pipe.OffsetStart != offset:
                pipe.OffsetStart = offset
                pipe.Proxy.execute(pipe)
        else:
            if pipe.OffsetEnd != offset:
                pipe.OffsetEnd = offset
                pipe.Proxy.execute(pipe)



from GeoFeature import GeoFeature


class Part(GeoFeature):
    """
    Author: Juergen Riegel (FreeCAD@juergen-riegel.net)
    Licence: LGPL
    This class handles document objects in Part
    """
    ...



#! python
# SPDX-License-Identifier: LGPL-2.1-or-later

# (c) 2006 Juergen Riegel

from . import template, templateModuleApp


class TemplateModule(template.ModelTemplate):
    def Generate(self):
        print("model.generateModel_Module.Generate()\n")
        App = templateModuleApp.TemplateModuleApp()
        App.outputDir = self.outputDir
        App.module = self.module
        App.Generate()



# Examples to generate trajectories out of shapes
import FreeCADGui as Gui

# getting selected edges from the selection and sort them
count = 0
FirstPos1 = None
FirstPos2 = None
LastPos2 = None
SortedEdgeList = []
for so in Gui.Selection.getSelectionEx():
    for edge in obj.SubObjects:
        if edge.Type != "Part::TopoShape":
            continue
        pos1 = edge.valueAt(0)
        pos2 = edge.valueAt(edge.Length)
        print(pos1, pos2)
        if count == 0:  # first edge
            FirstPos1 = pos1
            FirstPos2 = pos2
        elif count == 1:  # second edge
            continue
        else:  # the rest
            SortedEdgeList.append((pos1, pos2, edge))



# ***************************************************************************
# *   Copyright (c) 2025 Stefan Tröger <stefantroeger@gmx.net>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM histogram plot task panel"
__author__ = "Stefan Tröger"
__url__ = "https://www.freecad.org"

## @package task_post_histogram
#  \ingroup FEM
#  \brief task panel for post histogram plot

from PySide import QtCore, QtGui

import FreeCAD
import FreeCADGui

from . import base_fempostpanel
from femguiutils import extract_link_view as elv
from femguiutils import vtk_table_view

translate = FreeCAD.Qt.translate


class _TaskPanel(base_fempostpanel._BasePostTaskPanel):
    """
    The TaskPanel for editing properties of glyph filter
    """

    def __init__(self, vobj):
        super().__init__(vobj.Object)

        # data widget
        self.data_widget = QtGui.QWidget()
        hbox = QtGui.QHBoxLayout()
        self.data_widget.show_plot = QtGui.QPushButton()
        self.data_widget.show_plot.setText(translate("FEM", "Show Plot"))
        hbox.addWidget(self.data_widget.show_plot)
        self.data_widget.show_table = QtGui.QPushButton()
        self.data_widget.show_table.setText(translate("FEM", "Show Data"))
        hbox.addWidget(self.data_widget.show_table)
        vbox = QtGui.QVBoxLayout()
        vbox.addItem(hbox)
        vbox.addSpacing(10)

        extracts = elv.ExtractLinkView(self.obj, False, self)
        vbox.addWidget(extracts)

        self.data_widget.setLayout(vbox)
        self.data_widget.setWindowTitle(translate("FEM", "Histogram Data"))
        self.data_widget.setWindowIcon(FreeCADGui.getIcon(":/icons/FEM_PostHistogram.svg"))

        # histogram parameter widget
        self.view_widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/TaskPostHistogram.ui"
        )
        self.view_widget.setWindowTitle(translate("FEM", "Histogram View Settings"))
        self.view_widget.setWindowIcon(FreeCADGui.getIcon(":/icons/FEM_PostHistogram.svg"))

        self.__init_widgets()

        # form made from param and selection widget
        self.form = [self.data_widget, self.view_widget]

    # Setup functions
    # ###############

    def __init_widgets(self):

        # connect data widget
        self.data_widget.show_plot.clicked.connect(self.showPlot)
        self.data_widget.show_table.clicked.connect(self.showTable)

        # set current values to view widget
        viewObj = self.obj.ViewObject

        self.view_widget.Bins.setValue(viewObj.Bins)
        self._enumPropertyToCombobox(viewObj, "Type", self.view_widget.Type)
        self.view_widget.Cumulative.setChecked(viewObj.Cumulative)

        self.view_widget.Title.setText(viewObj.Title)
        self.view_widget.XLabel.setText(viewObj.XLabel)
        self.view_widget.YLabel.setText(viewObj.YLabel)

        self.view_widget.LegendShow.setChecked(viewObj.Legend)
        self._enumPropertyToCombobox(viewObj, "LegendLocation", self.view_widget.LegendPos)
        self.view_widget.BarWidth.setValue(viewObj.BarWidth)
        self.view_widget.HatchWidth.setValue(viewObj.HatchLineWidth)

        # connect callbacks
        self.view_widget.Bins.valueChanged.connect(self.binsChanged)
        self.view_widget.Type.activated.connect(self.typeChanged)
        self.view_widget.Cumulative.toggled.connect(self.comulativeChanged)

        self.view_widget.Title.editingFinished.connect(self.titleChanged)
        self.view_widget.XLabel.editingFinished.connect(self.xLabelChanged)
        self.view_widget.YLabel.editingFinished.connect(self.yLabelChanged)

        self.view_widget.LegendShow.toggled.connect(self.legendShowChanged)
        self.view_widget.LegendPos.activated.connect(self.legendPosChanged)
        self.view_widget.BarWidth.valueChanged.connect(self.barWidthChanged)
        self.view_widget.HatchWidth.valueChanged.connect(self.hatchWidthChanged)

    QtCore.Slot()

    def showPlot(self):
        self.obj.ViewObject.Proxy.show_visualization()

    QtCore.Slot()

    def showTable(self):

        # TODO: make data model update when object is recomputed
        data_model = vtk_table_view.VtkTableModel()
        data_model.setTable(self.obj.Table)

        dialog = QtGui.QDialog(self.data_widget)
        widget = vtk_table_view.VtkTableView(data_model)
        layout = QtGui.QVBoxLayout()
        layout.addWidget(widget)
        layout.setContentsMargins(0, 0, 0, 0)

        dialog.setLayout(layout)
        dialog.resize(1500, 900)
        dialog.show()

    QtCore.Slot(int)

    def binsChanged(self, bins):
        self.obj.ViewObject.Bins = bins

    QtCore.Slot(int)

    def typeChanged(self, idx):
        self.obj.ViewObject.Type = idx

    QtCore.Slot(bool)

    def comulativeChanged(self, state):
        self.obj.ViewObject.Cumulative = state

    QtCore.Slot()

    def titleChanged(self):
        self.obj.ViewObject.Title = self.view_widget.Title.text()

    QtCore.Slot()

    def xLabelChanged(self):
        self.obj.ViewObject.XLabel = self.view_widget.XLabel.text()

    QtCore.Slot()

    def yLabelChanged(self):
        self.obj.ViewObject.YLabel = self.view_widget.YLabel.text()

    QtCore.Slot(int)

    def legendPosChanged(self, idx):
        self.obj.ViewObject.LegendLocation = idx

    QtCore.Slot(bool)

    def legendShowChanged(self, state):
        self.obj.ViewObject.Legend = state

    QtCore.Slot(float)

    def barWidthChanged(self, value):
        self.obj.ViewObject.BarWidth = value

    QtCore.Slot(float)

    def hatchWidthChanged(self, value):
        self.obj.ViewObject.HatchLineWidth = value



from Base.Metadata import export, constmethod
from Base.PyObjectBase import PyObjectBase
from Part.App.Point import Point
from Part.App.TopoShape import TopoShape
from Part.App.TopoShapeEdge import TopoShapeEdge
from Part.App.TopoShapeFace import TopoShapeFace
from typing import overload, Final

@export(
    PythonName="Part.BRepOffsetAPI_MakeFilling",
    Include="BRepOffsetAPI_MakeFilling.hxx",
    Constructor=True,
    Delete=True,
)
class BRepOffsetAPI_MakeFilling(PyObjectBase):
    """
    N-Side Filling

    Author: Werner Mayer (wmayer[at]users.sourceforge.net)
    Licence: LGPL
    """

    def setConstrParam(
        self,
        *,
        Tol2d: float = 0.00001,
        Tol3d: float = 0.0001,
        TolAng: float = 0.01,
        TolCurv: float = 0.1
    ) -> None:
        """
        setConstrParam(Tol2d=0.00001, Tol3d=0.0001, TolAng=0.01, TolCurv=0.1)
        Sets the values of Tolerances used to control the constraint.
        """
        ...

    def setResolParam(
        self, *, Degree: int = 3, NbPtsOnCur: int = 15, NbIter: int = 2, Anisotropy: bool = False
    ) -> None:
        """
        setResolParam(Degree=3, NbPtsOnCur=15, NbIter=2, Anisotropy=False)
        Sets the parameters used for resolution.
        """
        ...

    def setApproxParam(self, *, MaxDeg: int = 8, MaxSegments: int = 9) -> None:
        """
        setApproxParam(MaxDeg=8, MaxSegments=9)
        Sets the parameters used to approximate the filling the surface
        """
        ...

    def loadInitSurface(self, face: TopoShapeFace) -> None:
        """
        loadInitSurface(face)
        Loads the initial surface.
        """
        ...

    @overload
    def add(self, Edge: TopoShapeEdge, Order: int, *, IsBound: bool = True) -> None: ...
    @overload
    def add(
        self, Edge: TopoShapeEdge, Support: TopoShapeFace, Order: int, *, IsBound: bool = True
    ) -> None: ...
    @overload
    def add(self, Support: TopoShapeFace, Order: int) -> None: ...
    @overload
    def add(self, Point: Point) -> None: ...
    @overload
    def add(self, U: float, V: float, Support: TopoShapeFace, Order: int) -> None: ...
    def add(self, **kwargs) -> None:
        """
        add(Edge, Order, IsBound=True)
        add(Edge, Support, Order, IsBound=True)
        add(Support, Order)
        add(Point)
        add(U, V, Support, Order)
        Adds a new constraint.
        """
        ...

    def build(self) -> None:
        """
        Builds the resulting faces.
        """
        ...

    def isDone(self) -> bool:
        """
        Tests whether computation of the filling plate has been completed.
        """
        ...

    @overload
    def G0Error(self) -> float: ...
    @overload
    def G0Error(self, arg: int) -> float: ...
    def G0Error(self, arg: int = 0) -> float:
        """
        G0Error([int])
        Returns the maximum distance between the result and the constraints.
        """
        ...

    @overload
    def G1Error(self) -> float: ...
    @overload
    def G1Error(self, arg: int) -> float: ...
    def G1Error(self, arg: int = 0) -> float:
        """
        G1Error([int])
        Returns the maximum angle between the result and the constraints.
        """
        ...

    @overload
    def G2Error(self) -> float: ...
    @overload
    def G2Error(self, arg: int) -> float: ...
    def G2Error(self, arg: int = 0) -> float:
        """
        G2Error([int])
        Returns the greatest difference in curvature between the result and the constraints.
        """
        ...

    def shape(self) -> TopoShape:
        """
        shape()
        Returns the resulting shape.
        """
        ...



#! python
# SPDX-License-Identifier: LGPL-2.1-or-later
# (c) 2008 Werner Mayer LGPL


import Part, math
import FreeCAD, FreeCADGui

App = FreeCAD
Gui = FreeCADGui
from FreeCAD import Base


def makeBottle(myWidth=50.0, myHeight=70.0, myThickness=30.0):
    aPnt1 = Base.Vector(-myWidth / 2.0, 0, 0)
    aPnt2 = Base.Vector(-myWidth / 2.0, -myThickness / 4.0, 0)
    aPnt3 = Base.Vector(0, -myThickness / 2.0, 0)
    aPnt4 = Base.Vector(myWidth / 2.0, -myThickness / 4.0, 0)
    aPnt5 = Base.Vector(myWidth / 2.0, 0, 0)

    aArcOfCircle = Part.Arc(aPnt2, aPnt3, aPnt4)
    aSegment1 = Part.LineSegment(aPnt1, aPnt2)
    aSegment2 = Part.LineSegment(aPnt4, aPnt5)

    aEdge1 = aSegment1.toShape()
    aEdge2 = aArcOfCircle.toShape()
    aEdge3 = aSegment2.toShape()
    aWire = Part.Wire([aEdge1, aEdge2, aEdge3])

    aTrsf = Base.Matrix()
    aTrsf.rotateZ(math.pi)  # rotate around the z-axis

    aMirroredWire = aWire.copy()
    aMirroredWire.transformShape(aTrsf)
    myWireProfile = Part.Wire([aWire, aMirroredWire])

    myFaceProfile = Part.Face(myWireProfile)
    aPrismVec = Base.Vector(0, 0, myHeight)
    myBody = myFaceProfile.extrude(aPrismVec)

    myBody = myBody.makeFillet(myThickness / 12.0, myBody.Edges)

    neckLocation = Base.Vector(0, 0, myHeight)
    neckNormal = Base.Vector(0, 0, 1)

    myNeckRadius = myThickness / 4.0
    myNeckHeight = myHeight / 10
    myNeck = Part.makeCylinder(myNeckRadius, myNeckHeight, neckLocation, neckNormal)
    myBody = myBody.fuse(myNeck)

    faceToRemove = 0
    zMax = -1.0

    for xp in myBody.Faces:
        surf = xp.Surface
        if type(surf) == Part.Plane:
            z = surf.Position.z
            if z > zMax:
                zMax = z
                faceToRemove = xp

    # This doesn't work for any reason
    myBody = myBody.makeThickness([faceToRemove], -myThickness / 50, 1.0e-3)
    myThreading = Part.makeThread(
        myNeckHeight / 10, myNeckRadius * 0.06, myHeight / 10, myNeckRadius * 0.99
    )
    myThreading.translate(Base.Vector(0, 0, myHeight))
    myCompound = Part.makeCompound([myBody, myThreading])

    return myCompound


def makeBoreHole():
    # create a document if needed
    if App.ActiveDocument is None:
        App.newDocument("Solid")

    Group = App.ActiveDocument.addObject("App::DocumentObjectGroup", "Group")
    Group.Label = "Bore hole"

    V1 = Base.Vector(0, 10, 0)
    V2 = Base.Vector(30, 10, 0)
    V3 = Base.Vector(30, -10, 0)
    V4 = Base.Vector(0, -10, 0)
    VC1 = Base.Vector(-10, 0, 0)
    C1 = Part.Arc(V1, VC1, V4)
    # and the second one
    VC2 = Base.Vector(40, 0, 0)
    C2 = Part.Arc(V2, VC2, V3)
    L1 = Part.LineSegment(V1, V2)
    # and the second one
    L2 = Part.LineSegment(V4, V3)
    S1 = Part.Shape([C1, C2, L1, L2])

    W = Part.Wire(S1.Edges)
    F = Part.makeFace(W)
    P = F.extrude(Base.Vector(0, 0, 5))

    # add objects with the shape
    Wire = Group.newObject("Part::Feature", "Wire")
    Wire.Shape = W
    Face = Group.newObject("Part::Feature", "Face")
    Face.Shape = F
    Prism = Group.newObject("Part::Feature", "Extrude")
    Prism.Shape = P

    c = Part.Circle(Base.Vector(0, 0, -1), Base.Vector(0, 0, 1), 2.0)
    w = Part.Wire(c.toShape())
    f = Part.makeFace(w)
    p = f.extrude(Base.Vector(0, 0, 7))
    P = P.cut(p)

    # add first borer
    Bore1 = Group.newObject("Part::Feature", "Borer_1")
    Bore1.Shape = p
    Hole1 = Group.newObject("Part::Feature", "Borer_Hole1")
    Hole1.Shape = P

    c = Part.Circle(Base.Vector(0, -11, 2.5), Base.Vector(0, 1, 0), 1.0)
    w = Part.Wire(c.toShape())
    f = Part.makeFace(w)
    p = f.extrude(Base.Vector(0, 22, 0))
    P = P.cut(p)

    # add second borer
    Bore2 = Group.newObject("Part::Feature", "Borer_2")
    Bore2.Shape = p
    Hole2 = Group.newObject("Part::Feature", "Borer_Hole2")
    Hole2.Shape = P

    App.ActiveDocument.recompute()

    # hide all objects except of the final one
    Gui.ActiveDocument.getObject(Wire.Name).hide()
    Gui.ActiveDocument.getObject(Face.Name).hide()
    Gui.ActiveDocument.getObject(Prism.Name).hide()
    Gui.ActiveDocument.getObject(Bore1.Name).hide()
    Gui.ActiveDocument.getObject(Hole1.Name).hide()
    Gui.ActiveDocument.getObject(Bore2.Name).hide()
    Gui.ActiveDocument.ActiveView.fitAll()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the Circle object."""
## @package circle
# \ingroup draftobjects
# \brief Provides the object code for the Circle object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftobjects.base import DraftObject
from draftutils import gui_utils
from draftutils import params


class Circle(DraftObject):
    """The Circle object"""

    def __init__(self, obj):
        super().__init__(obj, "Circle")

        _tip = QT_TRANSLATE_NOOP("App::Property", "Start angle of the arc")
        obj.addProperty("App::PropertyAngle", "FirstAngle", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP(
            "App::Property",
            "End angle of the arc (for a full circle, \
                give it same value as First Angle)",
        )
        obj.addProperty("App::PropertyAngle", "LastAngle", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Radius of the circle")
        obj.addProperty("App::PropertyLength", "Radius", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Create a face")
        obj.addProperty("App::PropertyBool", "MakeFace", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The area of this object")
        obj.addProperty("App::PropertyArea", "Area", "Draft", _tip, locked=True)

        obj.MakeFace = params.get_param("MakeFaceMode")

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(obj, vp_module="view_base", vp_class="ViewProviderDraft")

    def execute(self, obj):
        """This method is run when the object is created or recomputed."""
        if self.props_changed_placement_only():
            obj.positionBySupport()
            self.props_changed_clear()
            return

        import Part

        plm = obj.Placement

        shape = Part.makeCircle(
            obj.Radius.Value,
            App.Vector(0, 0, 0),
            App.Vector(0, 0, 1),
            obj.FirstAngle.Value,
            obj.LastAngle.Value,
        )

        if obj.FirstAngle.Value == obj.LastAngle.Value:
            shape = Part.Wire(shape)
            if getattr(obj, "MakeFace", True):
                shape = Part.Face(shape)

        obj.Shape = shape

        if hasattr(obj, "Area") and hasattr(shape, "Area"):
            obj.Area = shape.Area

        obj.Placement = plm
        obj.positionBySupport()
        self.props_changed_clear()

    def onChanged(self, obj, prop):
        self.props_changed_store(prop)


# Alias for compatibility with v0.18 and earlier
_Circle = Circle

## @}



from Base.Metadata import export, constmethod
from Part.App.ShapeFix.ShapeFix_Root import ShapeFix_Root
from typing import Final

@export(
    PythonName="Part.ShapeFix.Wire",
    Twin="ShapeFix_Wire",
    TwinPointer="ShapeFix_Wire",
    Include="ShapeFix_Wire.hxx",
    FatherInclude="Mod/Part/App/ShapeFix/ShapeFix_RootPy.h",
    Constructor=True,
)
class ShapeFix_Wire(ShapeFix_Root):
    """
    Class for fixing operations on wires

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    def init(self) -> None:
        """
        Initializes by wire, face, precision
        """
        pass

    def fixEdgeTool(self) -> None:
        """
        Returns tool for fixing wires
        """
        pass

    def clearModes(self) -> None:
        """
        Sets all modes to default
        """
        pass

    def clearStatuses(self) -> None:
        """
        Clears all statuses
        """
        pass

    def load(self) -> None:
        """
        Load data for the wire, and drops all fixing statuses
        """
        pass

    def setFace(self) -> None:
        """
        Set working face for the wire
        """
        pass

    def setSurface(self, surface: object, Placement: object = ...) -> None:
        """
        setSurface(surface, [Placement])
        Set surface for the wire
        """
        pass

    def setMaxTailAngle(self) -> None:
        """
        Sets the maximal allowed angle of the tails in radians
        """
        pass

    def setMaxTailWidth(self) -> None:
        """
        Sets the maximal allowed width of the tails
        """
        pass

    def isLoaded(self) -> None:
        """
        Tells if the wire is loaded
        """
        pass

    def isReady(self) -> None:
        """
        Tells if the wire and face are loaded
        """
        pass

    def numberOfEdges(self) -> None:
        """
        Returns number of edges in the working wire
        """
        pass

    def wire(self) -> None:
        """
        Makes the resulting Wire (by basic Brep_Builder)
        """
        pass

    def wireAPIMake(self) -> None:
        """
        Makes the resulting Wire (by BRepAPI_MakeWire)
        """
        pass

    def face(self) -> None:
        """
        Returns working face
        """
        pass

    def perform(self) -> None:
        """
        Iterates on subshapes and performs fixes
        """
        pass

    def fixReorder(self) -> None:
        """
        Performs an analysis and reorders edges in the wire
        """
        pass

    def fixSmall(self) -> None:
        """
        Applies fixSmall(...) to all edges in the wire
        """
        pass

    def fixConnected(self, num: int) -> None:
        """
        Applies fixConnected(num) to all edges in the wire
        Connection between first and last edges is treated only if
        flag ClosedMode is True
        If prec is -1 then maxTolerance() is taken.
        """
        pass

    def fixEdgeCurves(self) -> None:
        """
        Groups the fixes dealing with 3d and pcurves of the edges
        """
        pass

    def fixDegenerated(self) -> None:
        """
        Applies fixDegenerated(...) to all edges in the wire
        """
        pass

    def fixSelfIntersection(self) -> None:
        """
        Applies FixSelfIntersectingEdge(num) and
         FixIntersectingEdges(num) to all edges in the wire and
         FixIntersectingEdges(num1, num2) for all pairs num1 and num2
         and removes wrong edges if any
        """
        pass

    def fixLacking(self) -> None:
        """
        Applies FixLacking(num) to all edges in the wire
        Connection between first and last edges is treated only if
        flag ClosedMode is True
        If 'force' is False (default), test for connectness is done with
        precision of vertex between edges, else it is done with minimal
        value of vertex tolerance and Analyzer.Precision().
        Hence, 'force' will lead to inserting lacking edges in replacement
        of vertices which have big tolerances.
        """
        pass

    def fixClosed(self) -> None:
        """
        Fixes a wire to be well closed
        """
        pass

    def fixGaps3d(self, num: int) -> None:
        """
        Fixes gaps between ends of 3d curves on adjacent edges
        """
        pass

    def fixGaps2d(self, num: int) -> None:
        """
        Fixes gaps between ends of pcurves on adjacent edges
        """
        pass

    def fixSeam(self) -> None:
        """
        Fixes seam edges
        """
        pass

    def fixShifted(self) -> None:
        """
        Fixes edges which have pcurves shifted by whole parameter
        range on the closed surface
        """
        pass

    def fixNotchedEdges(self) -> None:
        """
        Fixes Notch edges.Check if there are notch edges in 2d and fix it
        """
        pass

    def fixGap3d(self, num: int) -> None:
        """
        Fixes gap between ends of 3d curves on num-1 and num-th edges
        """
        pass

    def fixGap2d(self, num: int) -> None:
        """
        Fixes gap between ends of pcurves on num-1 and num-th edges
        """
        pass

    def fixTails(self) -> None:
        """
        Fixes issues related to 'tails' in the geometry.
        Tails are typically small, undesired protrusions or deviations in the curves or edges that need correction.
        This method examines the geometry and applies corrective actions to eliminate or reduce the presence of tails.
        """
        pass
    ModifyTopologyMode: bool = ...
    """Mode for modifying topology of the wire"""

    ModifyGeometryMode: bool = ...
    """Mode for modifying geometry of vertexes and edges"""

    ModifyRemoveLoopMode: bool = ...
    """Mode for modifying edges"""

    ClosedWireMode: bool = ...
    """
    Mode which defines whether the wire
    is to be closed (by calling methods like fixDegenerated()
    and fixConnected() for last and first edges)
    """

    PreferencePCurveMode: bool = ...
    """
    Mode which defines whether the 2d 'True'
    representation of the wire is preferable over 3d one in the
    case of ambiguity in FixEdgeCurves
    """

    FixGapsByRangesMode: bool = ...
    """
    Mode which defines whether tool
    tries to fix gaps first by changing curves ranges (i.e.
    using intersection, extrema, projections) or not
    """

    FixReorderMode: bool = ...
    """
    Mode which performs an analysis and reorders edges in the wire using class WireOrder.
    Flag 'theModeBoth' determines the use of miscible mode if necessary.
    """

    FixSmallMode: bool = ...
    """Mode which applies FixSmall(num) to all edges in the wire"""

    FixConnectedMode: bool = ...
    """
    Mode which applies FixConnected(num) to all edges in the wire
    Connection between first and last edges is treated only if
    flag ClosedMode is True
    If 'prec' is -1 then MaxTolerance() is taken.
    """

    FixEdgeCurvesMode: bool = ...
    """
    Mode which groups the fixes dealing with 3d and pcurves of the edges.
    The order of the fixes and the default behaviour are:
    ShapeFix_Edge::FixReversed2d
    ShapeFix_Edge::FixRemovePCurve (only if forced)
    ShapeFix_Edge::FixAddPCurve
    ShapeFix_Edge::FixRemoveCurve3d (only if forced)
    ShapeFix_Edge::FixAddCurve3d
    FixSeam,
    FixShifted,
    ShapeFix_Edge::FixSameParameter
    """

    FixDegeneratedMode: bool = ...
    """
    Mode which applies FixDegenerated(num) to all edges in the wire
    Connection between first and last edges is treated only if
    flag ClosedMode is True
    """

    FixSelfIntersectionMode: bool = ...
    """
    Mode which applies FixSelfIntersectingEdge(num) and
    FixIntersectingEdges(num) to all edges in the wire and
    FixIntersectingEdges(num1, num2) for all pairs num1 and num2
    and removes wrong edges if any
    """

    FixLackingMode: bool = ...
    """
    Mode which applies FixLacking(num) to all edges in the wire
    Connection between first and last edges is treated only if
    flag ClosedMode is True
    If 'force' is False (default), test for connectness is done with
    precision of vertex between edges, else it is done with minimal
    value of vertex tolerance and Analyzer.Precision().
    Hence, 'force' will lead to inserting lacking edges in replacement
    of vertices which have big tolerances.
    """

    FixGaps3dMode: bool = ...
    """
    Mode which fixes gaps between ends of 3d curves on adjacent edges
    myPrecision is used to detect the gaps.
    """

    FixGaps2dMode: bool = ...
    """
    Mode whixh fixes gaps between ends of pcurves on adjacent edges
    myPrecision is used to detect the gaps.
    """

    FixReversed2dMode: bool = ...
    """Mode which fixes the reversed in 2d"""

    FixRemovePCurveMode: bool = ...
    """Mode which removePCurve in 2d"""

    FixAddPCurveMode: bool = ...
    """Mode which fixes addCurve in 2d"""

    FixRemoveCurve3dMode: bool = ...
    """Mode which fixes removeCurve in 3d """

    FixAddCurve3dMode: bool = ...
    """Mode which fixes addCurve in 3d"""

    FixSeamMode: bool = ...
    """Mode which fixes Seam """

    FixShiftedMode: bool = ...
    """Mode which fixes Shifted"""

    FixSameParameterMode: bool = ...
    """Mode which fixes sameParameter in 2d"""

    FixVertexToleranceMode: bool = ...
    """Mode which fixes VertexTolerence in 2d"""

    FixNotchedEdgesMode: bool = ...
    """Mode which fixes NotchedEdges in 2d"""

    FixSelfIntersectingEdgeMode: bool = ...
    """Mode which fixes SelfIntersectionEdge in 2d"""

    FixIntersectingEdgesMode: bool = ...
    """Mode which fixes IntersectingEdges in 2d"""

    FixNonAdjacentIntersectingEdgesMode: bool = ...
    """Mode which fixes NonAdjacentIntersectingEdges in 2d"""

    FixTailMode: bool = ...
    """Mode which fixes Tails in 2d"""
